"""

.. currentmodule:: basin3d.core.monitor

:synopsis: The BASIN-3D monitoring module
:module author: Val Hendrix <vhendrix@lbl.gov>

This module holds the logging functionality of BASIN-3D. It supports
logging the messages and their contexts

.. contents:: Contents
    :local:
    :backlinks: top

"""
import contextvars
import logging
import os
import uuid
from functools import wraps
from logging import config
from typing import Any, Callable, Dict, List, Optional, Union

import yaml

LOGGER_NAME = __name__
BASE_PATH = os.path.dirname(__file__)
LOG_CONFIG_PATH = os.path.join(BASE_PATH, "logging.yaml")

#: A unique identifier for a single request context. Use this
#: If you want to identify log records from a single context
#: This is a Context variable which is natively supported in asyncio
#: and are ready to be used without any extra configuration.
synthesis_id: contextvars.ContextVar = contextvars.ContextVar('synthesis_id')

#: A unique identifier for a  basin3d context. Use this
#: If you want to identify log records from a single context
#: This is a Context variable which is natively supported in asyncio
#: and are ready to be used without any extra configuration.
basin3d_where: contextvars.ContextVar = contextvars.ContextVar('basin3d_where')


class Basin3dLogger(logging.Logger):
    """
    Custom logger for adding realtime context information (basin3d_where)
    """

    def info(self, msg, *args, **kwargs):
        kwargs = self._add_extra(**kwargs)
        super().info(msg, *args, **kwargs)

    def error(self, msg, *args, **kwargs):
        kwargs = self._add_extra(**kwargs)
        super().error(msg, *args, **kwargs)

    def debug(self, msg, *args, **kwargs):
        kwargs = self._add_extra(**kwargs)
        super().debug(msg, *args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        kwargs = self._add_extra(**kwargs)
        super().warning(msg, *args, **kwargs)

    def critical(self, msg, *args, **kwargs):
        kwargs = self._add_extra(**kwargs)
        super().warning(msg, *args, **kwargs)

    def log(self, level, msg, *args, **kwargs):
        kwargs = self._add_extra(**kwargs)
        super().log(level, msg, *args, **kwargs)

    def _add_extra(self, **kwargs):
        """Add BASIN-3D extra, if exists """

        kwargs.setdefault('extra', dict())
        if "basin3d_where" not in kwargs['extra']:
            try:
                kwargs['extra']["basin3d_where"] = basin3d_where.get() or "*"
            except LookupError:
                kwargs['extra']["basin3d_where"] = "*"

        try:
            kwargs['extra']["synthesis_id"] = synthesis_id.get() or "*"
        except LookupError:
            kwargs['extra']["synthesis_id"] = "*"

        return kwargs


def get_logger(name: str = None) -> logging.Logger:
    """
    Get the basin3d logger for the specified name. Using this
    logger provides additional context for BASIN-3D synthesis.


    See :func:`logging.getLogger`

    :param: Name of the logger

    :return: The BASIN-3D Logger object
    """
    logging.setLoggerClass(Basin3dLogger)
    logger = logging.getLogger(name)

    return logger


def get_ctx_synthesis_id() -> Any:
    """
    Get the context for the monitoring synthesis_id
    :return:
    """
    try:
        return synthesis_id.get()
    except LookupError:
        pass

    return None


def set_ctx_synthesis_id() -> Any:
    """
    Set the context for the monitoring synthesis_id.
    :param synthesis_id: The request identifier
    """
    return synthesis_id.set(str(uuid.uuid1())[:8])


def get_ctx_basin3d_where() -> Any:
    """
    Get the context for the monitoring basin3d_where.
    """
    try:
        return basin3d_where.get()
    except LookupError:
        pass

    return None


def set_ctx_basin3d_where(where: Optional[Union[List, str]] = None) -> Union[Any, None]:
    """
    Set the context for the monitoring basin3d_where.
    :param where:
    """

    if isinstance(where, list):
        return basin3d_where.set(".".join(where))
    else:
        return basin3d_where.set(where)


def ctx_synthesis(func) -> Callable:
    """
    Decorator for setting synthesis id context

    :return: func
    """

    # Use of wraps makes sure that stack traces show the original
    # function name and not the wrapped one.
    @wraps(func)
    def func_wrapper(*args, **kwargs):

        s_token = set_ctx_synthesis_id()
        bw_token = set_ctx_basin3d_where()
        try:
            result = func(*args, **kwargs)
        finally:
            synthesis_id.reset(s_token)
            basin3d_where.reset(bw_token)
        return result
    return func_wrapper


def configure(log_config_path: str = None, **kwargs) -> Dict:
    """
    Load YAML python logging configuraiton file

    :param log_config_path: Path to the YAML file for configuring Python logging
    :returns: Logging configuration as dictionary

    **Keyword Args**
    Overwrite default logging config.

    + filters (dict)
    + formatters (dict)
    + handlers (dict)
    + loggers (dict)

    """
    # Load logging config
    logging.setLoggerClass(Basin3dLogger)
    log_config_path = log_config_path or LOG_CONFIG_PATH
    with open(f"{log_config_path}", "r") as f:
        # Expand any environment variables
        config_str = os.path.expandvars(f.read())

        # Load yaml yaml as dict
        config_file: Dict = yaml.load(config_str, Loader=yaml.Loader)
        config_file = _overwrite_config(config_file, config_overwrite=kwargs)
        config.dictConfig(config_file)

    return config_file


def _overwrite_config(config: Dict, config_overwrite: Dict) -> Dict:
    """
    Overwite the config with the specifed values
    :param config: The config to overwriete
    :param config_overwrite:
    :return:
    """

    if config_overwrite:
        # We are overwriting the config
        for key, value in config_overwrite.items():
            if key not in config.keys():
                config[key] = value
            else:
                if isinstance(value, dict):
                    _overwrite_config(config[key], config_overwrite[key])
                elif isinstance(value, (str, list)):
                    # Overwrite the configuration value
                    config[key] = value
                else:
                    raise Exception(f"Invalid config parameter {key}={value}. It must be a dict or string")

    return config


"""

.. currentmodule:: basin3d.core.translate

:synopsis: The BASIN-3D Translate functionality
:module author: Danielle Svehla Christianson <dschristianson@lbl.gov>

.. contents:: Contents
    :local:
    :backlinks: top

"""

from itertools import product, repeat
from typing import List, Optional, Union

from basin3d.core import monitor
from basin3d.core.schema.enum import MAPPING_DELIMITER, NO_MAPPING_TEXT
from basin3d.core.schema.query import QueryBase, QueryMeasurementTimeseriesTVP, QueryMonitoringFeature


logger = monitor.get_logger(__name__)


def _clean_query(translated_query: QueryBase) -> QueryBase:
    """
    Remove any NOT_SUPPORTED translations

    :param translated_query: the translated query that may have NOT_SUPPORTED translation results
    :return: cleaned query that has all NOT_SUPPORTED translations removed
    """
    for attr in translated_query.mapped_fields:
        attr_value = getattr(translated_query, attr)
        if attr_value and isinstance(attr_value, list):
            clean_list = [val for val in attr_value if val != NO_MAPPING_TEXT]
            unique_list = list(set(clean_list))
            setattr(translated_query, attr, unique_list)
        elif attr_value and attr_value == NO_MAPPING_TEXT:
            setattr(translated_query, attr, None)
    return translated_query


def _get_attr_types_in_compound_mappings(plugin_access) -> list:
    """
    Get all the compound mappings for a datasource if any exist

    :param plugin_access: plugin access
    :return: list of attribute types that are part of a compound mapping.
    """
    compound_mapping_attrs = set()

    attr_mapping_iterator = plugin_access.get_attribute_mappings()

    for attr_mapping in attr_mapping_iterator:
        if MAPPING_DELIMITER in attr_mapping.attr_type:
            compound_mapping_attrs.add(attr_mapping.attr_type)

    return list(compound_mapping_attrs)


def _get_attr_type_if_compound_mapping(plugin_access, attr_type: str) -> Optional[str]:
    """
    Return the compound attr_type str if the specified attr_type is part of a compound_mapping

    :param plugin_access:
    :param attr_type:
    :return: str if attr mapping is part of compound mapping
    """
    compound_mapping_str = None
    attr_type = attr_type.upper()
    attr_mapping_iterator = plugin_access.get_attribute_mappings(attr_type=attr_type)

    # look at the first element returned if there is one
    for attr_mapping in attr_mapping_iterator:
        if MAPPING_DELIMITER in attr_mapping.attr_type:
            compound_mapping_str = attr_mapping.attr_type
        # only need to look at the first attribute mapping returned
        break

    return compound_mapping_str


def _get_single_attr_types_in_compound_mappings(plugin_access, attr_type: str, include_specified_type: bool = False) -> list:
    """
    Return the attributes if attr_type is part of a compound mapping

    :param attr_type: the attribute type
    :param include_specified_type: bool, True = include in the return the specified attr_type. False: return the other attribute types that are part of the compound mapping.
    :return: list of attributes in the compound mapping
    """

    compound_mapping_attrs: List[str] = []

    compound_mapping_str = _get_attr_type_if_compound_mapping(plugin_access, attr_type)

    if not compound_mapping_str:
        return compound_mapping_attrs

    for attr in compound_mapping_str.split(MAPPING_DELIMITER):
        if attr == attr_type.upper() and not include_specified_type:
            continue
        compound_mapping_attrs.append(attr)

    return compound_mapping_attrs


def _is_translated_query_valid(datasource_id, query, translated_query) -> Optional[bool]:
    """
    Determine if the translated query is valid. A valid translated query has at least one datasoure vocabulary
    for any query field that was supecified (i.e., with a BASIN-3D vocab)

    :param datasource_id: the datasource id
    :param query: the original query
    :param translated_query: the translated query
    :return: boolean (True = valid translated query, False = invalid translated query) or None (translated query could not be assessed)
    """
    for field_type, field_list in zip(['mapped', 'prefixed'], [query.mapped_fields, query.prefixed_fields]):
        # loop thru kwargs
        for attr in field_list:
            translated_attr_value = getattr(translated_query, attr)
            b3d_attr_value = getattr(query, attr)
            msg_prefix = ''
            if field_type == 'mapped':
                msg_prefix = 'No vocabulary found for attribute {attr} with values: {b3d_attr_value}.'
            if isinstance(b3d_attr_value, list):
                b3d_attr_value = ', '.join(b3d_attr_value)
            if translated_attr_value and isinstance(translated_attr_value, list):
                # if list and all of list == NOT_SUPPORTED, False
                if all([x == NO_MAPPING_TEXT for x in translated_attr_value]):
                    logger.warning(f'Translated query for datasource {datasource_id} is invalid.{msg_prefix}')
                    return False
            elif translated_attr_value and isinstance(translated_attr_value, str):
                # if single NOT_SUPPORTED, False
                if translated_attr_value == NO_MAPPING_TEXT:
                    logger.warning(f'Translated query for datasource {datasource_id} is invalid.{msg_prefix}')
                    return False
            elif translated_attr_value:
                logger.warning(
                    f'Translated query for datasource {datasource_id} cannot be assessed. Translated value for {attr} is not expected type.')
                return None
    return True


def _order_mapped_fields(plugin_access, query_mapped_fields):
    """
    Order the mapped fields according to the order of the attributes in any compound mappings. And then handle the rest of the mappings.

    :param plugin_access:
    :param query_mapped_fields:
    :return:
    """
    query_mapped_fields_ordered = []

    # get list of compound mappings if any
    compound_mappings = _get_attr_types_in_compound_mappings(plugin_access)

    # If there are compound mappings...
    if compound_mappings:
        cm_fields = []
        # Split up the compound mappings, preserving the order of the attributes as specified in the plugin mapping file.
        # The order only matters relative to the individual compound mapping.
        for cm in compound_mappings:
            cm_attrs = cm.split(MAPPING_DELIMITER)
            cm_fields.extend([cm_attr.lower() for cm_attr in cm_attrs])
        # first loop thru the compound mapping fields
        for cm in cm_fields:
            # if the attribute is one of the mapped fields in this particular query
            if cm in query_mapped_fields:
                # add it to the ordered list
                query_mapped_fields_ordered.append(cm)
                # then remove it from the mapped field list
                query_mapped_fields.pop(query_mapped_fields.index(cm))
        # then, add any remaining non-compound fields
        query_mapped_fields_ordered.extend(query_mapped_fields)
    else:
        # if there are no compound mappings, then the order doesn't matter, just copy the mapped field list.
        query_mapped_fields_ordered = query_mapped_fields

    return query_mapped_fields_ordered


def _translate_mapped_query_attrs(plugin_access, query: Union[QueryMeasurementTimeseriesTVP, QueryMonitoringFeature]) -> QueryBase:
    """
    Translation functionality
    """
    query_mapped_fields = query.mapped_fields.copy()

    # if there are no mapped fields, return the query as is.
    if not query_mapped_fields:
        return query

    # order the query fields by any compound attributes
    query_mapped_fields_ordered = _order_mapped_fields(plugin_access, query_mapped_fields)

    for attr in query_mapped_fields_ordered:
        # if the attribute is specified, proceed to translate it
        # NOTE: looking in the translated_query which is mutable. As the translation occurs, translated query fields may change
        #       and the if statement may have different values for a given field during the loop.
        if getattr(query, attr):
            b3d_vocab = getattr(query, attr)

            if isinstance(b3d_vocab, str):
                ds_vocab = _translate_to_datasource_vocab(plugin_access, attr.upper(), b3d_vocab, query)
            else:
                ds_vocab = []
                for b3d_value in b3d_vocab:
                    # handle multiple values returned
                    ds_vocab.extend(_translate_to_datasource_vocab(plugin_access, attr.upper(), b3d_value, query))
            setattr(query, attr, ds_vocab)

            # look up whether the attr is part of a compound mapping
            compound_attrs = _get_single_attr_types_in_compound_mappings(plugin_access, attr)
            # if so: for any compound attrs, clear out the values in the synthesized query b/c search needs to be done on the coupled datasource_vocab
            for compound_attr in compound_attrs:
                setattr(query, compound_attr.lower(), None)

    # NOTE: always returns list for each mapped attr b/c multiple datasource vocab can be mapped to a single BASIN-3D vocab.
    return query


def _translate_prefixed_query_attrs(plugin_access, query: Union[QueryMeasurementTimeseriesTVP, QueryMonitoringFeature]) -> QueryBase:
    """

    :param plugin_access:
    :param query:
    :return:
    """
    def extract_id(identifer):
        """
        Extract the datasource identifier from the broker identifier
        :param identifer:
        :return:
        """
        if identifer:
            site_list = identifer.split("-")
            identifer = identifer.replace("{}-".format(site_list[0]), "", 1)  # The datasource id prefix needs to be removed
        return identifer

    id_prefix = plugin_access.datasource.id_prefix

    for attr in query.prefixed_fields:
        attr_value = getattr(query, attr)
        if attr_value:

            # if the value is a string
            if isinstance(attr_value, str):
                translated_value = extract_id(attr_value)
                if translated_value == attr_value:
                    translated_value = NO_MAPPING_TEXT

            # otherwise assume it is a list
            else:
                translated_value = [extract_id(x) for x in attr_value if x.startswith("{}-".format(id_prefix))]

            setattr(query, attr, translated_value)

    return query


def _translate_to_datasource_vocab(plugin_access, attr_type: str, basin3d_vocab: str, b3d_query) -> list:
    """
    Find the datasource vocabulary(ies) for the specified datasource, attribute type, BASIN-3D vocabulary, and full query that may specify other attributes.
    Because multiple datasource vocabularies can be mapped to the same BASIN-3D vocabulary, the return is a list of the datasource vocabs.

    :param plugin_access: plugin access
    :param attr_type: the attribute type
    :param basin3d_vocab: the BASIN-3D vocabulary
    :param b3d_query: either a QueryBase class or subclass object, or a dictionary
    :return: list of the datasource vocabularies
    """
    # convert attr_type to uppercase
    attr_type = attr_type.upper()

    # is the attr_type part of a compound mapping?
    compound_mapping_attrs = _get_single_attr_types_in_compound_mappings(plugin_access, attr_type, include_specified_type=True)

    b3d_vocab_combo_str = [basin3d_vocab]

    # if compound_mapping, find all the relevant value combos given the query
    if compound_mapping_attrs:
        b3d_vocab_filter_lists = []  # list to hold lists of specified filters, one for each attr

        # loop thru each of the compound mapping attributes, build a list of lists of query combos
        for attr in compound_mapping_attrs:
            attr_value = None

            # by default: match any number of characters excepting a new line for the attribute
            # replace this value below if a value for the attribute is specified in the query
            filter_values: List[str] = ['.*']
            # if the attr is the attr_type, set the filter to the specified vocab
            if attr == attr_type:
                filter_values = [basin3d_vocab]
            elif issubclass(b3d_query.__class__, QueryBase) and hasattr(b3d_query, attr.lower()):
                attr_value = getattr(b3d_query, attr.lower())
            elif isinstance(b3d_query, dict) and attr.lower() in b3d_query.keys():
                attr_value = b3d_query.get(attr.lower())

            # if there is a value, replace the default value
            if attr_value:
                filter_values = attr_value
                # if the values are a str, change it to a list
                if isinstance(attr_value, str):
                    filter_values = attr_value.split(',')

            # append the filter list to the main list
            b3d_vocab_filter_lists.append(filter_values)

        # create a list of sets contain the combinations of filter options for each attr
        b3d_vocab_combo_sets = list(product(*b3d_vocab_filter_lists))

        # change each set into a str for search
        b3d_vocab_combo_str = [MAPPING_DELIMITER.join(v) for v in b3d_vocab_combo_sets]

    ds_vocab: List[str] = []
    no_match_list: List[str] = []

    # Loop thru the list of vocabulary string combos to search the attribute mapping database
    for basin3d_vocab_str in b3d_vocab_combo_str:
        results_iterator = plugin_access.get_attribute_mappings(attr_type=attr_type,
                                                                attr_vocab=basin3d_vocab_str, from_basin3d=True)
        query_results = []

        # there may be more than one datasource variable mapped to the specified BASIN-3D vocab
        # Loop thru the results and collect all of them.
        for qr in results_iterator:
            query_results.append(qr.datasource_vocab)

        # If there were indeed mappings, add them to the main ds_vocab list and move on
        if query_results:
            ds_vocab.extend(query_results)
            continue

        # if not result for the combo, add it to the no_match_list
        no_match_list.append(basin3d_vocab_str)

    if not ds_vocab:
        ds_vocab = [NO_MAPPING_TEXT]

    if no_match_list:
        if compound_mapping_attrs:
            attr_type = ':'.join(compound_mapping_attrs)
        logger.info(f'Datasource "{plugin_access.datasource.id}" did not have matches for attr_type '
                    f'{attr_type} and BASIN-3D vocab {", ".join(no_match_list)}.')

    return ds_vocab


def get_datasource_mapped_attribute(plugin_access, attr_type, datasource_vocab):
    """
    Get the :class:`basin3d.core.models.MappedAttribute` object(s) for the specified attribute type and datasource attribute vocab(s)

    :param plugin_access: plugin_access
    :param attr_type: attribute type
    :param datasource_vocab: datasource attribute vocabulary
    :return: a single or list of `basin3d.core.models.MappedAttribute` objects
    """

    if isinstance(datasource_vocab, str):
        return plugin_access.get_datasource_attribute_mapping(attr_type, datasource_vocab)

    elif isinstance(datasource_vocab, list):
        return list(map(plugin_access.get_datasource_attribute_mapping, repeat(attr_type), datasource_vocab))


def translate_attributes(plugin_access, mapped_attrs, **kwargs):
    """
    Translate datasource vocabularies to BASIN-3D vocabularies via
    :class:`basin3d.core.models.AttributeMapping` objects specified in the datasource plugin
    for specified model attributes (i.e., those that have mappings).
    This function is used by data model objects during object creation. See :class:`basin3d.core.models`.

    :param plugin_access: plugin_access
    :param mapped_attrs: a model's mapped attributes
    :param kwargs: the full set of a model's attributes
    :return: kwargs: the model's attributes, including the translated attributes
    """

    # copy the kwargs be able to loop thru the original while modifying the actual for compound mappings
    kwargs_orig = kwargs.copy()

    for attr in mapped_attrs:
        if attr in kwargs_orig:
            datasource_vocab = kwargs[attr]
            attr_mapping = get_datasource_mapped_attribute(plugin_access, attr_type=attr.upper(), datasource_vocab=datasource_vocab)
            kwargs[attr] = attr_mapping

            # If the attr is part of a compound mapping and the compound attr is not part of the kwargs, set it.
            cm_attrs = _get_single_attr_types_in_compound_mappings(plugin_access, attr)
            if cm_attrs:
                for cm_attr in cm_attrs:
                    if cm_attr.lower() not in kwargs:
                        cm_attr_mapping = get_datasource_mapped_attribute(plugin_access, attr_type=cm_attr.upper(), datasource_vocab=datasource_vocab)
                        kwargs[cm_attr.lower()] = cm_attr_mapping

    return kwargs


def translate_query(plugin_access, query: Union[QueryMeasurementTimeseriesTVP, QueryMonitoringFeature]) -> QueryBase:
    """
    Translate BASIN-3D vocabulary specified in a query to the datasource vocabularies defined by :class:`basin3d.core.models.AttributeMapping` objects specified in the datasource plugin.

    :param plugin_access: plugin access
    :param query: query to be translated
    :return: translated query as a :class:`basin3d.core.schema.query.QueryBase` object
    """
    translated_query = query.copy()
    _translate_mapped_query_attrs(plugin_access, translated_query)
    _translate_prefixed_query_attrs(plugin_access, translated_query)
    is_valid_translated_query = _is_translated_query_valid(plugin_access.datasource.id, query, translated_query)

    if is_valid_translated_query:
        translated_query.is_valid_translated_query = is_valid_translated_query
        _clean_query(translated_query)

    return translated_query



"""

.. currentmodule:: basin3d.core.plugin

:platform: Unix, Mac
:synopsis: BASIN-3D ``DataSource`` plugin classes
:module author: Val Hendrix <vhendrix@lbl.gov>
:module author: Danielle Svehla Christianson <dschristianson@lbl.gov>

.. contents:: Contents
    :local:
    :backlinks: top


"""
from types import MethodType
from typing import Dict

from basin3d.core import monitor
from basin3d.core.catalog import CatalogTinyDb
from basin3d.core.models import DataSource
from basin3d.core.schema.enum import FeatureTypeEnum

logger = monitor.get_logger(__name__)


def get_feature_type(feature_type, return_format="enum"):
    """
    Return the feature type if exists in the request
    :param feature_type:
    :param return_format: "enum" (default) = the FeatureTypeEnum enum,
    otherwise return the text version
    :return: the feature_type in the format specified, None if none exists
    """
    if feature_type:
        for k, v in FeatureTypeEnum.__members__.items():
            ft = v.lower()
            if ft == feature_type.lower():
                if return_format == "enum":
                    return k
                else:
                    return v
    return None


class DataSourcePluginAccess:
    """
    Metaclass for DataSource plugin views.  The should be registered in a subclass of
    :class:`basin3d.plugins.DataSourcePluginPoint` in attribute `plugin_access_classes`.
    """
    def __init__(self, datasource: DataSource, catalog: CatalogTinyDb):
        """

        :param datasource: The data source object
        :param catalog: The Plugin Catalog
        """
        self._datasource = datasource
        self._catalog = catalog

    @property
    def datasource(self):
        return self._datasource

    def get_datasource_attribute_mapping(self, attr_type, attr_vocab):
        """
        Get attribute mapping for the specified attribute type and datasource attribute vocab

        :param attr_type: attribute type
        :param attr_vocab: datasource attribute vocabulary
        :return: a `basin3d.models.AttributeMapping` object
        """
        return self._catalog.find_datasource_attribute_mapping(self.datasource.id, attr_type, attr_vocab)

    def get_attribute_mappings(self, attr_type=None, attr_vocab=None, from_basin3d=False):
        """
        General purpose search for attribute mappings from datasource or BASIN-3D vocabularies.
        Returns an iterator that yields `basin3d.models.MappedAttribute` objects that match the criteria
        :param attr_type:
        :param attr_vocab:
        :param from_basin3d:
        :return:
        """
        return self._catalog.find_attribute_mappings(self.datasource.id, attr_type, attr_vocab, from_basin3d)


def basin3d_plugin(cls):
    """Register a BASIN-3D plugin"""

    # TODO add any validation here for class structure
    PluginMount.plugins[cls.get_meta().id] = cls
    return cls


def _basin3d_plugin_access(plugin_class, synthesis_model_class, access_type):
    """Decorator for registering model access"""

    def _inner(func):

        plugin_access_classes = getattr(plugin_class, 'plugin_access_classes', None)
        if not plugin_access_classes:
            setattr(plugin_class, 'plugin_access_classes', [])
            plugin_access_classes = getattr(plugin_class, 'plugin_access_classes')

        class_name = f"{plugin_class.__module__}.{plugin_class.get_meta().id}{synthesis_model_class.__name__}Access"

        new_class = None
        for plugin_access in plugin_access_classes:
            if plugin_access.__name__ in class_name:
                new_class = plugin_access

        if new_class:
            setattr(new_class, access_type, MethodType(func, new_class))
        else:
            new_class = type(class_name, (DataSourcePluginAccess,),
                             {"synthesis_model_class": synthesis_model_class, access_type: func})
            plugin_access_classes.append(new_class)

        return func

    return _inner


class DataSourcePluginPoint:
    """
    Base class for DataSourcePlugins.
    """

    def __init__(self, catalog: CatalogTinyDb):
        """
        Instantiate the plugin with the catalog
        :param catalog:
        """

        meta = self.get_meta()
        datasource_id = getattr(meta, 'id')
        self.datasource = DataSource(id=datasource_id,
                                     id_prefix=self.get_id_prefix(),
                                     name=self.get_name(),
                                     location=self.get_location(),
                                     credentials={})

        self.access_classes = {}
        plugin_access_classes = getattr(self, 'plugin_access_classes', None)
        if plugin_access_classes:
            for access_class in plugin_access_classes:
                access_point = access_class(self.get_datasource(), catalog)
                self.access_classes[access_point.synthesis_model_class] = access_point

    @classmethod
    def get_meta(cls):
        """
        Gets the DataSourceMeta internal class that should be defined by subclasses.
        Raises an error if it is not found
        :return:
        """
        meta = getattr(cls, 'DataSourceMeta', None)

        if not meta:
            raise ValueError("Must define inner class DatasSourceMeta for {}".format(cls))
        return meta

    def get_datasource(self):
        """
        Get the `basin3d.models.DataSource` record for the subclass of this Plugin.
        # :return:
        """
        return self.datasource

    def get_plugin_access(self):
        """
        Get the defined plugin_access_classes from the subclass.  These should be defined in
        `DataSourceMeta.plugin_view_subclass`. If not, an error is thrown
        :return:
        """
        return self.access_classes

    @classmethod
    def get_id_prefix(cls):
        """
        Get the defined id prefix
        """
        meta = cls.get_meta()
        return getattr(meta, 'id_prefix')

    @classmethod
    def get_name(cls):
        """
        Get the defined id prefix
        """
        meta = cls.get_meta()
        return getattr(meta, 'name')

    @classmethod
    def get_location(cls):
        """
        Get the defined id prefix
        """
        meta = cls.get_meta()
        return getattr(meta, 'location')

    @classmethod
    def get_feature_types(cls):
        """
        Get the defined feature types
        :return:
        """
        return getattr(cls, 'feature_types', None)


class PluginMount(type):
    """
    The idea for the Simple Plugin Framework comes from a post
    by Marty Alchin on January 10, 2008 about Django

    See: http://martyalchin.com/2008/jan/10/simple-plugin-framework/

    it is under CC-BY-SA 3.0 US License
    (https://creativecommons.org/licenses/by-sa/3.0/us/)

    Plugin classes that extend this will register themselves as
    soon as they are loaded

    """
    plugins: Dict[DataSourcePluginPoint, str] = {}



from enum import Enum
from dataclasses import dataclass
from datetime import date
import os
import pandas as pd
import pathlib
from typing import Dict, List, Optional, Iterator, Union, Tuple

from basin3d import monitor
from basin3d.core.schema.enum import FeatureTypeEnum
from basin3d.core.schema.query import QueryMeasurementTimeseriesTVP, QueryMonitoringFeature
from basin3d.core.models import AbsoluteCoordinate, AltitudeCoordinate, Coordinate, DepthCoordinate, \
    GeographicCoordinate, HorizontalCoordinate, RepresentativeCoordinate, \
    MeasurementTimeseriesTVPObservation, MonitoringFeature, \
    TimeMetadataMixin, TimeValuePair, ResultListTVP
from basin3d.core.plugin import DataSourcePluginAccess, DataSourcePluginPoint, basin3d_plugin
from basin3d.core.types import SpatialSamplingShapes

logger = monitor.get_logger(__name__)

ESSDIVE_DATASETS_PATH = os.environ.get('ESSDIVE_DATASETS_PATH', None)


# ToDo: This should not be an enum class, rework in future.
class HydroRFTerms(Enum):
    # ESS-DIVE Hydrologic Monitoring Reporting Format
    # https://github.com/ess-dive-community/essdive-hydrologic-monitoring

    lat = 'Latitude'  # Required
    long = 'Longitude'  # Required
    depth = 'Depth'
    sensor_depth = 'Sensor_Depth'
    depth_reference = 'Depth_Reference'  # Look for depth units
    elevation = 'Elevation'
    sensor_elevation = 'Sensor_Elevation'
    elevation_reference = 'Elevation_Reference'  # Look for reference and datum
    loc_id = 'Site_ID'
    site_name = 'Site_Name'
    water_name = 'Water_Name'
    sensor_id = 'Sensor_ID'  # Not officially part of the terminology but is allowed in the HydroRF_Instructions.md
    loc_type = 'Site_Type'  # Search for river = HORIZONTAL_PATH, lake | groundwater = WATER_BODY, ecosystem = ECOSYSTEM, well = POINT
    utc_offset = 'UTC_Offset'
    install_method_id = 'InstallationMethod_ID'
    install_method_desc = 'InstallationMethod_Description'

    # time variables
    date_time = 'DateTime'
    date_time_start = 'DateTime_Start'
    date_time_end = 'DateTime_End'
    date_time_vars = [date_time, date_time_start, date_time_end]

    # observation_terms and units -- variables commented out are not supported yet.
    variables = {'Depth_To_Water': {'rf_options': ['meters', 'centimeters'], 'convert': 'to_m'},  # m
                 'Gage_Height': {'rf_options': ['meters', 'centimeters'], 'convert': 'to_m'},  # m
                 # 'Sensor_Depth': {'rf_options': ['meters', 'centimeters'], 'convert': 'to_m'},  # m
                 # 'Sensor_Elevation': {'rf_options': ['meters_above_mean_sea_level_NAVD88'], 'convert': None},
                 # 'Sensor_Pressure_Vented': {'rf_options': ['kilopascals', 'meters_of_H2O'], 'convert': 'tbd'},
                 # 'Sensor_Pressure_Unvented': {'rf_options': ['kilopascals', 'meters_of_H2O'], 'convert': 'tbd'},
                 'Water_Depth': {'rf_options': ['meters', 'centimeters'], 'convert': 'to_m'},  # m
                 'Water_Surface_Elevation': {'rf_options': ['meters_above_mean_sea_level_NAVD88'], 'convert': None},
                 'Electrical_Conductivity': {'rf_options': ['microsiemens_per_centimeter', 'millisiemens_per_centimeter'], 'convert': 'to_micro'},  # uS/cm
                 'Specific_Conductance': {'rf_options': ['microsiemens_per_centimeter', 'millisiemens_per_centimeter'], 'convert': 'to_micro'},  # uS/cm
                 'pH': {'rf_options': ['pH'], 'convert': None},
                 'Dissolved_Oxygen': {'rf_options': ['milligrams_per_liter'], 'convert': None},
                 # 'Dissolved_Oxygen_Saturation': {'rf_options': ['percent_saturation'], 'convert': None},
                 'Water_Temperature': {'rf_options': ['degree_celsius'], 'convert': None}}

    # location terms
    location_terms = [lat, long, depth, depth_reference, elevation, elevation_reference, loc_id,
                      site_name, water_name, sensor_id, loc_type, utc_offset, install_method_id]
    loc_numeric_type = [lat, long, depth, elevation]

    # unique location terms
    # following approach in which a unique location is at the surface for the monitoring_feature (mf),
    #    adding depths / elevation to mf desc when the information is easily accessible (i.e., in the locations files)
    # mf objects associated with measurement objects will have a singular depth
    unique_loc_terms = [loc_id]

    # equivalent terms -- note elevation is defined as sensor only
    loc_names = [site_name, water_name]

    # header info
    header_row_char = '#'
    format_row_text = '# HeaderRows_Format:'
    header_delimiter = ';'
    column_header = 'Column_Header'
    units = 'Unit'

    missing_value_str = 'N/A'
    missing_value_numeric = -9999

    @classmethod
    def get_attr_name(cls, target_value: str) -> Optional[str]:
        for a_var, a_value in cls.__members__.items():
            if target_value == a_value.value:
                return a_var
        return None


@dataclass
class locInfo:
    ds_id: str
    ds_pid: str
    loc_id: str
    lat: Optional[float] = None
    long: Optional[float] = None
    depth: Optional[List[float]] = None
    sensor_depth: Optional[float] = None
    depth_reference: Optional[str] = None
    elevation: Optional[List[float]] = None
    sensor_elevation: Optional[float] = None
    elevation_reference: Optional[str] = None
    site_name: Optional[str] = None
    water_name: Optional[str] = None
    sensor_id: Optional[str] = None
    utc_offset: Optional[int] = None
    install_method_id: Optional[str] = None
    install_method_desc: Optional[str] = None
    loc_type: str = 'POINT'  # Future Search for river = HORIZONTAL_PATH, lake | groundwater = WATER_BODY, ecosystem = ECOSYSTEM, well = POINT


class UnitHandler:
    """ Class to handle specific unit conversions for the Hydrological Monitoring RF"""

    def __init__(self):
        self.fuzzy_match_store = {
            'centimeters': ['cm', 'centimeter'],
            'meters': ['m', 'meters'],
            'microsiemens_per_centimeter': ['uS/cm', 'uS / cm'],
            'millisiemens_per_centimeter': ['mS/cm', 'mS / cm'],
            'degree_celsius': ['deg C', 'deg_C', 'C'],
            'kilopascals': ['kPa'],
            'percent_saturation': ['%', 'percent'],
            'meters_above_mean_sea_level_NAVD88': ['meters', 'm', 'meter']
        }

    @staticmethod
    def cm_to_m(unit: str) -> float:
        if unit == "centimeters":
            return 0.01
        return 1

    @staticmethod
    def milli_to_micro(unit: str) -> int:
        if 'milli' in unit:
            return 1000
        return 1

    # ToDo: handle conversion between kPa and meters H2O depending on tbd BASIN-3D variable and units

    def match_unit(self, rf_units: List[str], unit: str) -> Optional[str]:
        """
        Try to match unit
        :param rf_units:
        :param unit:
        :return:
        """
        unit_match = None
        for rf_unit in rf_units:
            unit_syn_list = self.fuzzy_match_store.get(rf_unit, [])
            for unit_syn in unit_syn_list:
                if unit_syn == unit:
                    unit_match = rf_unit
                    return unit_match
        return unit_match

    def convert_value(self, rf_unit: str, convert_type: Optional[str]) -> Union[float, int]:
        if convert_type == 'to_m':
            return self.cm_to_m(rf_unit)
        elif convert_type == 'to_micro':
            return self.milli_to_micro(rf_unit)
        return 1


def _ess_dive_datasets_handler(dirpath: Optional[str] = ESSDIVE_DATASETS_PATH) -> dict:
    """
    Read ESS-DIVE Datasets directory and build dictionary with dataset info for each dataset

    Note: dataset info is updated by _build_locations_store if it is called after this handler method

    :param dirpath:
    :return:
    """
    datasets: Dict[str, Dict] = {}

    if dirpath is None:
        logger.warning('ESSDIVE_DATASETS_PATH not specified or configured in environmental variables.')
        return datasets

    # make path a Path object so works on linux and PC
    essdive_path = pathlib.Path(dirpath)

    for dataset in os.listdir(essdive_path):
        ds_path = essdive_path / dataset

        if not pathlib.Path.is_dir(ds_path):
            continue

        # assumes that directory name has format: <datasetID>-<datasetName>-pid-<datasetUUID>
        ds_full_name = dataset.split('-pid-')[0]
        ds_pieces = ds_full_name.split('-')
        ds_id = ds_pieces[0]
        ds_pid = dataset.split('-pid-')[-1]
        ds_name = '-'.join(ds_pieces[1:])

        ds_key = (ds_id, ds_pid)

        datasets.setdefault(ds_key, {}).update({  # type: ignore[arg-type]
            'pid': ds_pid,
            'id': ds_id,
            'name': ds_name,
            'locations': ds_path / 'locations',
            'data': ds_path / 'data',
            'dataset': ds_path
        })

    return datasets


def _list_csv_files(dir_path: str) -> List:
    """
    Return list of csv files in specified directory
    :param dir_path:
    :return:
    """
    dir_list = os.listdir(dir_path)
    return [f for f in dir_list if f.endswith('.csv')]


def _read_csv_to_pandas_df(filepath: pathlib.Path, columns_only: bool = False) -> Tuple[List, pd.DataFrame]:
    """
    Read a csv file into a pandas dataframe
    :param filepath:
    :param columns_only: bool, return the first couple rows so that the columns can be assessed efficiently
    :return:
    """
    with open(filepath, mode='r', encoding='utf-8') as f:

        # assess the number of header rows
        header_lines = []
        for line in f:
            if line.startswith('#'):
                header_lines.append(line)
            else:
                break

        f.seek(0)
        header_ct = len(header_lines)
        if columns_only:
            nrows = 2
        else:
            nrows = None

        pdf = pd.read_csv(f, header=header_ct, skipinitialspace=True, keep_default_na=False, on_bad_lines='warn', nrows=nrows)

    return header_lines, pdf


def _create_loc_key(ds_id: str, pd_row: Optional[pd.Series] = None, loc_dict: dict = {}) -> tuple:
    """
    Create the location key
    :param ds_id: Dataset id
    :param pd_row: a pandas dataframe row
    :param loc_dict: a location dictionary object
    :return: a tuple consisting of dataset id and location id
    """
    callable_obj = pd_row
    if callable_obj is None:
        callable_obj = loc_dict

    key_elements = [ds_id]

    for term in HydroRFTerms.unique_loc_terms.value:
        if term in callable_obj.keys():
            key_elements.append(callable_obj[term])
            continue
        key_elements.append('None')

    return tuple(key_elements)


def _make_lat_long_key(lat: float, long: float) -> str:
    """
    Make a location id from the lat, long information
    :param lat:
    :param long:
    :return:
    """
    return f'LAT{str(lat)}_LON{str(long)}'


def _has_lat_long_terms(term_list: list) -> bool:
    """
    Helper function to check if lat / long terms are present
    :param term_list:
    :return:
    """
    if HydroRFTerms.lat.value in term_list and HydroRFTerms.long.value in term_list:
        return True
    return False


def _extract_ds_id(mf_list: list) -> list:
    """
    Extract dataset identifier from monitoring feature list
    :param mf_list:
    :return:
    """
    ds_id_list = set()
    for mf_id in mf_list:
        mf_id_pieces = mf_id.split('-')
        ds_id_list.add(mf_id_pieces[0])

    return list(ds_id_list)


def _parse_data_file_header_rows(header_rows: List[str]) -> dict:
    """
    Parse a data file's header rows following the RF
    :param header_rows:
    :return:
    """
    header_store: Dict = {}

    loc_header_idx = []
    loc_header_terms = []

    for idx, row in enumerate(header_rows):
        if idx == 0:
            continue
        if idx == 1:
            if not row.startswith(HydroRFTerms.format_row_text.value):
                logger.error(f'Second header row in data file does not start with "{HydroRFTerms.format_row_text.value}"')
                return header_store
            format_terms_str = row.replace(HydroRFTerms.format_row_text.value, '').replace('\n', '').rstrip(',')
            format_terms = format_terms_str.split(HydroRFTerms.header_delimiter.value)
            format_terms = [x.strip() for x in format_terms]
            if not format_terms or format_terms[0] != HydroRFTerms.column_header.value or format_terms[1] != HydroRFTerms.units.value:
                logger.error(f'Second header row could not be parsed with delimiter {HydroRFTerms.header_delimiter.value} '
                             f'or the first header term was not {HydroRFTerms.column_header.value} '
                             f'or the second header term was not {HydroRFTerms.units.value}')
                return header_store
            for i, term in enumerate(format_terms):
                if term in HydroRFTerms.location_terms.value:
                    loc_header_idx.append(i)
                    loc_header_terms.append(term)
            header_store = {
                'loc_header_terms': loc_header_terms,
                'loc_header_idx': loc_header_idx,
                'loc_column_info': {},
                'variable_column_info': {},
                'time_column_info': {}
            }
            continue

        column_info_str = row.replace(HydroRFTerms.header_row_char.value, '').replace('\n', '').rstrip(',')
        column_info = column_info_str.split(HydroRFTerms.header_delimiter.value)
        column_info = [x.strip() for x in column_info]
        column_name = column_info[0]
        column_name_pieces = column_name.split('_')
        potential_var_col_name = '_'.join(column_name_pieces[:-1])
        sensor_id, variable_name = None, None
        if column_name in HydroRFTerms.location_terms.value:
            store_dict = header_store['loc_column_info']
        elif column_name in HydroRFTerms.variables.value.keys() or potential_var_col_name in HydroRFTerms.variables.value.keys():
            store_dict = header_store['variable_column_info']
            variable_name = column_name
            if potential_var_col_name in HydroRFTerms.variables.value.keys():
                sensor_id = column_name_pieces[-1]
                variable_name = potential_var_col_name
        elif column_name in HydroRFTerms.date_time_vars.value:
            store_dict = header_store['time_column_info']
        else:
            logger.info(f'header term {column_name} is not defined in RF. Skipping.')
            continue
        store_dict_element: Dict = store_dict.setdefault(column_name, {'unit': column_info[1]})
        for i, row_idx in enumerate(loc_header_idx):
            store_dict_element[loc_header_terms[i]] = column_info[row_idx]
        if variable_name:
            store_dict_element.update({HydroRFTerms.sensor_id.value: sensor_id, 'variable': variable_name})

    return header_store


def _check_columns_for_loc_info(data_pdf: pd.DataFrame, header_store: Dict):
    """
    Because data file header rows are not required to be a 1:1 match with the column names,
    look for additional loc info in the columns and update the header_store.
    :param data_pdf:
    :param header_store:
    """
    data_cols = list(data_pdf.columns)
    loc_col_info = header_store['loc_column_info']

    for col_name in data_cols:
        if col_name not in HydroRFTerms.location_terms.value:
            continue
        if col_name not in loc_col_info.keys():
            loc_col_info[col_name] = {}


def _get_queried_variable_info(variable_col_info: Dict, queried_observed_properties: List) -> Tuple[Dict, List, bool]:
    """
    Extract helper info to determine which of the queried observed properties are present in the data variables.
    This is not a 1:1 match on column names so extra handling is required.
    :param variable_col_info:
    :param queried_observed_properties:
    :return:
    """
    queried_col_info = {}
    var_set = set()
    var_list = []
    for col_name, col_info in variable_col_info.items():
        var = col_info.get('variable')
        if var and var in queried_observed_properties:
            var_set.add(var)
            var_list.append(var)
            queried_col_info.update({col_name: col_info})

    has_multiple_same_var = False
    var_set_list = list(var_set)
    if len(var_list) > len(var_set_list):
        has_multiple_same_var = True

    return queried_col_info, var_set_list, has_multiple_same_var


def _has_valid_datetime_col(time_info_store: dict) -> bool:
    """
    Confirm that a valid datetime column is present.
    :param time_info_store:
    :return:
    """
    # if no valid date time variables, return False
    if not time_info_store:
        return False

    # if date_time_end is present, then date_time_start must also be
    if HydroRFTerms.date_time_end.value in time_info_store.keys():
        if HydroRFTerms.date_time_start.value in time_info_store.keys() and HydroRFTerms.date_time.value not in time_info_store.keys():
            return True
        return False

    # If both date_time and date_time_start, return False
    if HydroRFTerms.date_time.value in time_info_store.keys() and HydroRFTerms.date_time_start.value in time_info_store.keys():
        return False

    # Otherwise if either date_time or date_time_start is present, then OK.
    return True


def _get_primary_date_time_col(time_info_store: dict) -> str:
    """
    BASIN-3D does not handle start and end timestamps in the time value pairs object yet.
    Choose a the timestamp start column if both start and end are reported
    :param time_info_store:
    :return:
    """
    if HydroRFTerms.date_time_start.value in time_info_store.keys():
        return HydroRFTerms.date_time_start.value

    return HydroRFTerms.date_time.value


def _build_locations_store(datasets: dict, mf_locations: list) -> Dict[Tuple, locInfo]:
    """
    Compile location information from the datasets give the requested mf_locations if any.
    :param datasets:
    :param mf_locations:
    :return:
    """

    # -------- Method internal functions

    def create_loc_info(ds_id: str, ds_pid: str, loc_id: str, loc_callable_obj: Union[pd.Series, dict]) -> locInfo:
        """
        Create an locinfo object instance
        :param ds_id:
        :param ds_pid:
        :param loc_id:
        :param loc_callable_obj:
        :return:
        """
        loc_info = locInfo(ds_id=ds_id, ds_pid=ds_pid, loc_id=loc_id)
        loc_info = update_loc_info(loc_info, callable_obj=loc_callable_obj)
        return loc_info

    def update_loc_info(loc_info_obj: locInfo, callable_obj: Union[pd.Series, dict]):
        """
        Update a locInfo object
        :param loc_info_obj:
        :param callable_obj:
        :return:
        """
        for term in HydroRFTerms.location_terms.value:
            if term != HydroRFTerms.loc_id.value and term in callable_obj.keys():
                attr_name = HydroRFTerms.get_attr_name(term)
                if attr_name is None:
                    logger.warning(f'Attribute name for term {term} is {attr_name}. This should never happen. Contact developer.')
                    continue

                attr_value = callable_obj[term]

                if isinstance(attr_value, pd.Series):
                    attr_value = attr_value.iloc[0]

                if term in HydroRFTerms.loc_numeric_type.value:
                    try:
                        attr_value = float(attr_value)
                    except Exception as e:
                        logger.warning(f'{e}')
                        continue
                # ToDo: potentially pass thru missing values -- RF requires missing values be report for empty values if the field is included but not enforcing that.
                if attr_value and (attr_value == HydroRFTerms.missing_value_numeric.value if isinstance(attr_value, float) or isinstance(attr_value, int) else False or
                                   attr_value == HydroRFTerms.missing_value_str.value if isinstance(attr_value, str) else False):
                    continue

                if term == HydroRFTerms.depth.value or term == HydroRFTerms.elevation.value:
                    value = getattr(loc_info_obj, attr_name)
                    if value:
                        setattr(loc_info_obj, attr_name, value.append(attr_value))
                    else:
                        setattr(loc_info_obj, attr_name, [attr_value])
                    continue
                setattr(loc_info_obj, attr_name, attr_value)
        return loc_info_obj

    def identify_loc_files(csv_files: List[str], loc_dir: pathlib.Path) -> Tuple[Optional[str], Optional[str]]:
        """
        Identify the filenames of the location files within the dataset.
        :param csv_files:
        :param loc_dir:
        :return:
        """
        install_methods_file = None
        site_id_file = None

        for f in csv_files:
            f_lower = f.lower()
            if not install_methods_file and 'installation' in f_lower and 'method' in f_lower:
                install_methods_file = f
                continue
            elif not site_id_file:
                try:
                    _, pdf = _read_csv_to_pandas_df(loc_dir / f)
                    if HydroRFTerms.loc_id.value in pdf.columns:
                        site_id_file = f
                        continue
                except Exception as e:
                    logger.info(f'MESSAGE: {e}')
                    continue

        return install_methods_file, site_id_file

    def parse_location_file(locations_store, loc_pdf, loc_file, is_additional, additional_pdf, additional_file,
                            im_has_site_id, ds_id, ds_pid, id_term, mf_locations, is_site_id_file=False):
        """
        Parse either the InstallationMethods or supplemental locations file.
        :param locations_store:
        :param loc_pdf:
        :param loc_file:
        :param is_additional:
        :param additional_pdf:
        :param additional_file:
        :param im_has_site_id:
        :param ds_id:
        :param ds_pid:
        :param id_term:
        :param mf_locations:
        :param is_site_id_file:
        :return:
        """
        nrows = loc_pdf.shape[0]
        for idx in range(0, nrows):
            loc_row = loc_pdf.iloc[idx, ]

            lat = loc_row[HydroRFTerms.lat.value]
            long = loc_row[HydroRFTerms.long.value]

            if not isinstance(lat, float) or not isinstance(long, float):
                logger.info(f'{ds_id}: {loc_file} row {idx} does not have valid lat {lat} or long {long} values. Skipping')
                continue

            # loc_id = f'LAT{str(lat)}_LONG{str(long)}'
            loc_id = _make_lat_long_key(lat, long)
            if im_has_site_id or is_site_id_file:
                loc_id = loc_row[id_term]

            # if locations are specified, and the current location is not specified, skip it
            if mf_locations and f'{ds_id}-{loc_id}' not in mf_locations:
                continue

            loc_key = _create_loc_key(ds_id, pd_row=loc_row)
            if loc_key in locations_store.keys():
                logger.warning(f'Row {idx} of {loc_file} duplicated info for {loc_key}. Skipping')
                continue

            loc_info = create_loc_info(ds_id, ds_pid, loc_id, loc_row)

            if im_has_site_id and is_additional:
                site_id_row = additional_pdf[additional_pdf[id_term] == loc_id]
                if site_id_row.shape[0] == 1:
                    update_loc_info(loc_info, site_id_row)
                    continue
                # ToDo: consider adding handling for multiple depths reported for different variables in the IM
                logger.info(f'{ds_id}: {additional_file} as multiple entries for {loc_id}. Only one entry allowed per {loc_id}. Skipping location info in file.')

            locations_store.update({loc_key: loc_info})

    def update_ds_info(ds_store: dict, im_has_lat_long: bool, im_has_site_id: bool, s_has_lat_long: bool,
                       id_term: Optional[str], has_id_in_column: Optional[bool], has_lat_long_in_header: Optional[bool],
                       has_lat_long_in_column: Optional[bool], install_methods_file: Optional[str] = None):
        """
        Update the dataset store for where the location information was found.

        :param ds_store:
        :param im_has_lat_long:
        :param im_has_site_id:
        :param s_has_lat_long:
        :param id_term:
        :param has_id_in_column:
        :param has_lat_long_in_header:
        :param has_lat_long_in_column:
        :return:
        """
        ds_store.update({
            'location_info': {
                'im_has_lat_long': im_has_lat_long,
                'im_has_site_id': im_has_site_id,
                's_has_lat_long': s_has_lat_long,
                'id_term': id_term,
                'has_id_in_column': has_id_in_column,
                'has_lat_long_in_header': has_lat_long_in_header,
                'has_lat_long_in_column': has_lat_long_in_column},
            'install_methods_file': install_methods_file})

    # ---- Main functionality for method

    # Because locations can occur multiple times across several files,
    #    keep track of what has been yielded with fields: <site_id>, <lat>, <long>)
    #    Assumptions:
    #    1. A dataset is consistent in that data files won't have differing locations fields in different places
    #    2. Elevation and depth references are consistent
    #
    #    Choices:
    #    1. A sensor is not a location so replicated sensors at the same location will
    #       not be treated as separate locations. This maybe tricky in the pandas df column name in a view.
    #    2. ID is formed by lat / long if site ID is not provided.
    #    3. Site IDs can have different depths -- going to treat them as the same location in the mf calls,
    #       and a specific location in the timeseries calls. Same column name issue in views as #1

    locations_store: Dict = {}

    ds_ids = _extract_ds_id(mf_locations)

    # ToDo: maybe go get DOI from essdive using pid?
    for ds_key, ds_info in datasets.items():
        ds_id, ds_pid = ds_key

        if ds_ids and ds_id not in ds_ids:
            continue

        loc_dir = ds_info.get('locations')
        loc_files = _list_csv_files(loc_dir)
        # Find the InstallationMethods file and see if there is an additional file that has the Site_ID term
        # If the additional location file does not have Site_ID, cannot support it.
        install_methods_file, site_id_file = identify_loc_files(loc_files, loc_dir)

        if not install_methods_file:
            logger.info(f'{ds_id}-{ds_pid}: No Installation Methods file detected. Cannot parse dataset.')
            continue

        loc_terms_install_id = []
        loc_terms_site_id = []

        id_term: Optional[str] = None
        im_has_site_id = False
        im_is_additional = False

        im_has_lat_long = False
        s_has_lat_long = False
        s_is_additional = False

        im_header_rows, im_pdf = _read_csv_to_pandas_df(loc_dir / install_methods_file)
        im_pdf_columns = list(im_pdf.columns)

        for col_term in im_pdf_columns:
            if col_term in HydroRFTerms.location_terms.value:
                loc_terms_install_id.append(col_term)

        if HydroRFTerms.install_method_id.value not in loc_terms_install_id and HydroRFTerms.install_method_desc.value not in loc_terms_install_id:
            logger.info(f'{ds_id}-{ds_pid} Installation Methods file does not have required fields. Cannot parse.')
            continue

        if _has_lat_long_terms(loc_terms_install_id):
            im_has_lat_long = True

        if not im_has_lat_long and len(loc_terms_install_id) > 2:
            im_is_additional = True

        im_pdf = im_pdf[loc_terms_install_id]

        if HydroRFTerms.loc_id.value in loc_terms_install_id:
            id_term = HydroRFTerms.loc_id.value
            im_has_site_id = True

        s_pdf: Optional[pd.DataFrame] = None
        if site_id_file:

            s_header_rows, s_pdf = _read_csv_to_pandas_df(loc_dir / site_id_file)
            s_pdf_columns = list(s_pdf.columns)

            for col_term in s_pdf_columns:
                if col_term in HydroRFTerms.location_terms.value:
                    loc_terms_site_id.append(col_term)

            if _has_lat_long_terms(loc_terms_site_id):
                if im_has_lat_long:
                    logger.info(f'{ds_id}-{ds_pid}: both Installation Methods and additional locations file have lat / long. Using Install Methods.')
                else:
                    s_has_lat_long = True
                    id_term = HydroRFTerms.loc_id.value

            if not s_has_lat_long and len(loc_terms_site_id) > 1:
                s_is_additional = True

            s_pdf = s_pdf[loc_terms_site_id]

        # ToDo: consider maybe expanding to handle depths in data files
        # ToDo: consider using Sensor_ID (sort of part of RF) as connector -- wait for use case

        has_id_in_column: Optional[bool] = None
        has_lat_long_in_header: Optional[bool] = None
        has_lat_long_in_column: Optional[bool] = None

        # If the lat / long terms are in the Installation Methods file, generate locations from that file.
        if im_has_lat_long:
            logger.info(f'{ds_id}-{ds_pid}: {install_methods_file}: Found lat / long in Installation Methods file. Assuming units / datum is Decimal degrees WGS84.')

            parse_location_file(locations_store=locations_store, loc_pdf=im_pdf, loc_file=install_methods_file,
                                is_additional=s_is_additional, additional_pdf=s_pdf, additional_file=site_id_file,
                                im_has_site_id=im_has_site_id, id_term=id_term, ds_id=ds_id, ds_pid=ds_pid, mf_locations=mf_locations)

            update_ds_info(ds_info, im_has_lat_long, im_has_site_id, s_has_lat_long, id_term,
                           has_id_in_column, has_lat_long_in_header, has_lat_long_in_column, install_methods_file)

        # Otherwise, if the lat / long terms are in the Site ID file, generate locations from this file
        elif s_has_lat_long:
            logger.info(f'{ds_id}-{ds_pid}: {site_id_file}: Found lat / long in additional locations file. Assuming units / datum is Decimal degrees WGS84.')

            parse_location_file(locations_store=locations_store, loc_pdf=s_pdf, loc_file=site_id_file,
                                is_additional=im_is_additional, additional_pdf=im_pdf, additional_file=install_methods_file,
                                im_has_site_id=im_has_site_id, id_term=id_term, ds_id=ds_id, ds_pid=ds_pid,
                                mf_locations=mf_locations, is_site_id_file=True)

            update_ds_info(ds_info, im_has_lat_long, im_has_site_id, s_has_lat_long, id_term,
                           has_id_in_column, has_lat_long_in_header, has_lat_long_in_column, install_methods_file)

        # Finally, if not in either location file, look for lat / long + locations info in the data files
        else:

            data_dir = ds_info.get('data')
            data_files = _list_csv_files(data_dir)
            has_id_in_column = False
            has_lat_long_in_header = False
            has_lat_long_in_column = False

            if not data_files:
                ds_path = ds_info.get('ds_path')
                logger.warning(f'Dataset {ds_id}-{ds_pid} does not have location or data files in {ds_path}. Cannot find locations info.')
                continue

            # Inspect the first data file to determine how to parse the remaining files
            data_file = data_files[0]
            data_header_rows, data_pdf = _read_csv_to_pandas_df(data_dir / data_file, columns_only=True)

            # parse the header rows
            header_store = _parse_data_file_header_rows(data_header_rows)
            _check_columns_for_loc_info(data_pdf, header_store)

            if not header_store:
                logger.warning(f'{ds_id}-{ds_pid}: Upon initial data file inspection, {data_file} has malformed header. Cannot parse dataset.')
                continue

            if _has_lat_long_terms(header_store['loc_header_terms']):
                has_lat_long_in_header = True
            elif _has_lat_long_terms(list(header_store['loc_column_info'].keys())):
                has_lat_long_in_column = True

            if HydroRFTerms.loc_id.value in header_store['loc_column_info'].keys():
                has_id_in_column = True
                id_term = HydroRFTerms.loc_id.value

            update_ds_info(ds_info, im_has_lat_long, im_has_site_id, s_has_lat_long, id_term,
                           has_id_in_column, has_lat_long_in_header, has_lat_long_in_column, install_methods_file)

            # -------- start the logic --------

            # if lat / long is in both the header and column, won't deal with it.
            if has_lat_long_in_header and has_lat_long_in_column:
                logger.info(f'{ds_id}-{ds_pid}: Data files based on initial file inspection of {data_file} has lat / long info in both the header and the columns. Cannot parse dataset.')
                continue

            # if lat / long in the header rows, loop through the rows that describe the columns if they are not location terms
            elif has_lat_long_in_header:
                logger.info(f'{ds_id}-{ds_pid}: Found lat / long in initial data file header inspection. Assuming units / datum is Decimal degrees WGS84.')

                for data_file in data_files:
                    data_header_rows, data_pdf = _read_csv_to_pandas_df(data_dir / data_file)

                    # parse the header rows
                    header_store = _parse_data_file_header_rows(data_header_rows)
                    _check_columns_for_loc_info(data_pdf, header_store)

                    if not header_store:
                        logger.warning(f'{ds_id}-{ds_pid}: Data file {data_file} has malformed header. Cannot parse. Skipping file.')
                        continue

                    if not _has_lat_long_terms(header_store['loc_header_terms']):
                        logger.warning(f'{ds_id}-{ds_pid}: Cannot find lat / long terms in {data_file}. Cannot parse. Skipping file.')
                        continue

                    for col_name, col_info in header_store['variable_column_info'].items():
                        try:
                            lat = float(col_info.get(HydroRFTerms.lat.value))
                            long = float(col_info.get(HydroRFTerms.long.value))
                        except Exception:
                            lat = None
                            long = None

                        if not isinstance(lat, float) or not isinstance(long, float):
                            logger.info(f'{ds_id}-{ds_pid}: {data_file}: {col_name} does not have valid lat {lat} or long {long} values. Skipping')
                            continue

                        loc_id = _make_lat_long_key(lat, long)
                        im_id = col_info.get(HydroRFTerms.install_method_id.value)
                        im_row: Optional[pd.Series] = None

                        try:
                            im_row = im_pdf[im_pdf[HydroRFTerms.install_method_id.value] == im_id]
                            if im_row.shape[0] > 1:
                                logger.info(f'{ds_id}-{ds_pid}: {install_methods_file} as multiple entries for {im_id}. '
                                            f'Only one entry allowed. Skipping integration with Install Methods for {col_name}.')
                                continue
                        except Exception:
                            pass

                        if HydroRFTerms.loc_id.value in header_store['loc_header_terms']:
                            loc_id = col_info[HydroRFTerms.loc_id.value]
                        elif im_has_site_id and im_row:
                            loc_id = im_row[HydroRFTerms.loc_id.value]
                        elif HydroRFTerms.loc_id.value in header_store['loc_column_info']:
                            logger.info(f'{ds_id}-{ds_pid}: lat / long info is in the header rows and {HydroRFTerms.loc_id.value} '
                                        f'is in the columns and cannot be utilized. Continuing with Lat/long identifier.')

                        loc_key = _create_loc_key(ds_id, loc_dict={HydroRFTerms.loc_id.value: loc_id,
                                                                   HydroRFTerms.lat.value: lat,
                                                                   HydroRFTerms.long.value: long})

                        if mf_locations and f'{ds_id}-{loc_id}' not in mf_locations:
                            continue

                        if loc_key in locations_store.keys():
                            logger.warning(f'{ds_id}-{ds_pid}: {data_file}: Loc info for {col_name} has duplicated info for {loc_key}. Skipping')
                            continue

                        loc_info = create_loc_info(ds_id, ds_pid, loc_id, col_info)

                        if im_is_additional:
                            update_loc_info(loc_info, im_row)

                        locations_store.update({loc_key: loc_info})

            # lat / long is in the columns, loop thru the pandas data frame rows
            elif has_lat_long_in_column:
                logger.info(f'{ds_id}: Found lat / long in initial data file column inspection. Assuming units / datum is Decimal degrees WGS84.')

                for data_file in data_files:
                    data_header_rows, data_pdf = _read_csv_to_pandas_df(data_dir / data_file)

                    # parse the header rows
                    header_store = _parse_data_file_header_rows(data_header_rows)
                    _check_columns_for_loc_info(data_pdf, header_store)

                    if not header_store:
                        logger.warning(f'{ds_id}-{ds_pid}: Data file {data_file} has malformed header. Cannot parse. Skipping file.')
                        continue

                    columns_store = header_store['loc_column_info']

                    if not _has_lat_long_terms(list(columns_store.keys())):
                        logger.warning(f'{ds_id}-{ds_pid}: Data file {data_file} does not have lat / long columns as expected. Skipping file.')
                        continue

                    data_pdf = data_pdf[list(columns_store.keys())]
                    group_cols = [HydroRFTerms.lat.value, HydroRFTerms.long.value]

                    if has_id_in_column and HydroRFTerms.loc_id.value not in columns_store.keys():
                        logger.info(f'{ds_id}-{ds_pid}: {data_file} Site ID should be available as a column and it is not. Skipping file.')
                        continue
                    elif has_id_in_column:
                        group_cols.append(HydroRFTerms.loc_id.value)

                    unique_groups = list(data_pdf.groupby(group_cols).indices.keys())

                    if not unique_groups:
                        logger.info(f'{ds_id}-{ds_pid}: {data_files} Could not find unique location groups among file rows. Skipping')
                        continue

                    for loc_group in unique_groups:
                        lat = loc_group[0]
                        long = loc_group[1]

                        if not isinstance(lat, float) or not isinstance(long, float):
                            logger.info(f'{ds_id}-{ds_pid}: {data_file} does not have valid lat {lat} or long {long} values. Skipping')
                            continue

                        loc_id = _make_lat_long_key(lat, long)
                        if has_id_in_column:
                            loc_id = loc_group[2]

                        loc_dict = {HydroRFTerms.loc_id.value: loc_id,
                                    HydroRFTerms.lat.value: lat,
                                    HydroRFTerms.long.value: long}

                        loc_key = _create_loc_key(ds_id, loc_dict=loc_dict)

                        if mf_locations and f'{ds_id}-{loc_id}' not in mf_locations:
                            continue

                        # We might expect the location to be in multiple files so not reporting log message.
                        if loc_key in locations_store.keys():
                            continue

                        loc_info = create_loc_info(ds_id, ds_pid, loc_id, loc_dict)
                        locations_store.update({loc_key: loc_info})

            else:
                logger.info(f'{ds_id}-{ds_pid}: Cannot find lat / long info in data files based on initial file inspection of {data_file}. Cannot parse dataset.')

    return locations_store


def _build_loc_store_index(ds_id: str, loc_store: dict) -> dict:
    """
    Build an index to the locations store
    This is a helper index b/c the key is a dataset id, location id tuple.
    :param ds_id:
    :param loc_store:
    :return:
    """
    key_index = {}
    for key_tuple in loc_store.keys():
        key_ds_id, loc_id = key_tuple
        if key_ds_id == ds_id:
            key_index.update({loc_id: key_tuple})

    return key_index


def _make_mf_object(plugin_access: DataSourcePluginAccess, mf_info: locInfo,
                    is_mf_query: bool = False, additional_data_info: dict = {}) -> Optional[MonitoringFeature]:
    """
    Construct a monitoring feature object instance
    :param plugin_access:
    :param mf_info:
    :param additional_data_info:
    :return:
    """

    def configure_elev_depth_fields(term: str, mf_info: locInfo, value, is_mf_query: bool) -> Tuple[dict, Optional[str]]:
        """
        Handle elevation and depth fields
        :param term: RF Depth or Elevation term
        :param mf_info: location info object from locations store
        :param value: the value acquired from the data columns
        :return:
        """
        elev_depth: Dict = {'value': None}

        # get the value for the specified term; if there is not and also no explicit value, return
        term_value = getattr(mf_info, term)
        if not value and not term_value:
            return elev_depth, ''

        # try to get the term reference
        ref = getattr(mf_info, f'{term}_reference')

        # is_mf_query, make a string of depths
        if is_mf_query:
            if value:
                term_value = [value]
            txt = ', '.join(str(v) for v in term_value)
            txt = f' Observations at known {term}s: {txt} {ref}. These values may differ from those reported within the data files.'
            return elev_depth, txt

        depth_height_value = value
        if not value and term_value:
            depth_height_value = term_value[0]

        if depth_height_value == HydroRFTerms.missing_value_numeric.value:
            return elev_depth, ''

        elev_depth['value'] = depth_height_value

        ref_datum = None
        ref_unit = None

        unit = DepthCoordinate.DISTANCE_UNITS_METERS
        datum = AltitudeCoordinate.DATUM_NAVD88
        if term == 'depth':
            datum = DepthCoordinate.DATUM_LOCAL_SURFACE

        if not ref:
            logger.info(f'{mf_info.ds_id}-{mf_info.ds_pid}: {term.capitalize()}_Reference not reported with {term} values. Assuming {datum} per RF.')
            ref_datum = datum
            ref_unit = unit
        elif term == 'elevation' and datum in ref:
            ref_datum = datum
        elif term == 'depth' and 'ground surface' in ref:
            ref_datum = datum

        if ref and unit in ref:
            ref_unit = unit
        elif ref:
            logger.info(f'{mf_info.ds_id}-{mf_info.ds_pid}: {term.capitalize()}_Reference does not contain standard units')
            if DepthCoordinate.DISTANCE_UNITS_FEET in ref:
                ref_unit = DepthCoordinate.DISTANCE_UNITS_FEET

        if ref_datum:
            elev_depth.update({'datum': ref_datum})
        if ref_unit:
            elev_depth.update({'distance_units': ref_unit})

        return elev_depth, ''

    # set up depth and elev info. If comes from data files and it separate from the location info, get it from the data parsing.
    depth, elev = None, None
    if additional_data_info:
        depth = additional_data_info.get('depth')
        elev = additional_data_info.get('elev')

    depth_info, depth_txt = configure_elev_depth_fields('depth', mf_info, depth, is_mf_query)
    elev_info, elev_txt = configure_elev_depth_fields('elevation', mf_info, elev, is_mf_query)

    mf_name = mf_info.site_name
    if not mf_name:
        mf_name = mf_info.water_name
    if not mf_name:
        mf_name = mf_info.loc_id

    coordinates = Coordinate(
            absolute=AbsoluteCoordinate(
                horizontal_position=GeographicCoordinate(
                    **{'latitude': mf_info.lat,
                       'longitude': mf_info.long,
                       'datum': HorizontalCoordinate.DATUM_WGS84,
                       'units': GeographicCoordinate.UNITS_DEC_DEGREES})))

    elev_str = ''
    if elev_info.get('value'):
        coordinates.absolute.vertical_extent = [AltitudeCoordinate(**elev_info)]
        elev_str = f'-ELEV{str(elev_info.get("value"))}'

    depth_str = ''
    if depth_info.get('value'):
        coordinates.representative = RepresentativeCoordinate(vertical_position=DepthCoordinate(**depth_info))
        depth_str = f'-DEPTH{str(depth_info.get("value"))}'

    add_desc = ''
    sensor_str = ''
    if mf_info.sensor_id:
        add_desc = f' Sensor_ID: {mf_info.sensor_id}.'
        sensor_str = f'-SENSOR{mf_info.sensor_id}'
    if mf_info.install_method_desc:
        add_desc = f'{add_desc} Installation Method Description: {mf_info.install_method_desc}.'

    if additional_data_info:
        col_unit = additional_data_info.get('col_unit')
        col_rf_unit = additional_data_info.get('col_rf_unit')
        extra_unit_txt = ''
        if col_unit != col_rf_unit:
            convert_txt = ''
            if additional_data_info.get('conversion_factor') != 1:
                convert_txt = f' with conversion factor {additional_data_info.get("conversion_factor")}'
            extra_unit_txt = f' that was changed to RF unit {col_rf_unit}{convert_txt}'
        add_desc = (f'{add_desc} Data file: {additional_data_info.get("data_file")}; '
                    f'Column name: {additional_data_info.get("col_name")} '
                    f'with units {col_unit}{extra_unit_txt}.')

    desc = f'ESSDIVE dataset: {mf_info.ds_id}; pid: {mf_info.ds_pid}.{elev_txt}{depth_txt}{add_desc}'

    monitoring_feature = MonitoringFeature(
        plugin_access,
        id=f'{mf_info.ds_id}-{mf_info.loc_id}{elev_str}{depth_str}{sensor_str}',
        name=mf_name,
        description=desc,
        feature_type=mf_info.loc_type,
        shape=SpatialSamplingShapes.SHAPE_POINT,
        observed_properties=[],
        coordinates=coordinates
    )

    return monitoring_feature


def _parse_data_file(plugin_access: DataSourcePluginAccess, data_dir: pathlib.Path, data_file: str,
                     observed_property: list, mf_locations: list, start_date: date, end_date: Optional[date],
                     id_term: Optional[str], id_in_header: bool, id_in_col: bool, id_ref_im: bool, has_sensor_id: bool,
                     ds_id: str, ds_pid: str, im_ref_store: dict, loc_store_index: dict, locations_store: dict) -> Iterator:
    """
    Parse data file that follows the variations of the RF.

    :param plugin_access:
    :param data_dir:
    :param data_file:
    :param observed_property:
    :param mf_locations:
    :param start_date:
    :param end_date:
    :param id_term:
    :param id_in_header:
    :param id_in_col:
    :param id_ref_im:
    :param has_sensor_id:
    :param ds_id:
    :param ds_pid:
    :param ds_info:
    :param im_ref_store:
    :param loc_store_index:
    :param locations_store:
    :return:
    """
    data_header_rows, data_pdf = _read_csv_to_pandas_df(data_dir / data_file)

    # parse the header rows and check column names for any location terms
    header_store = _parse_data_file_header_rows(data_header_rows)
    _check_columns_for_loc_info(data_pdf, header_store)

    if not header_store:
        logger.warning(f'{ds_id}-{ds_pid}: {data_file} has malformed header. Cannot parse file.')
        yield None

    # No datetime variable, skip.
    if not _has_valid_datetime_col(header_store.get('time_column_info', {})):
        logger.info(f'{ds_id}-{ds_pid}: Data file {data_file} does not have valid date / time column names. Skipping.')
        yield None

    # Filter by the start date. Skip file if no date matches query dates.
    time_primary_var = _get_primary_date_time_col(header_store['time_column_info'])

    # First make sure date time values are date or datetime
    try:
        for time_col in header_store['time_column_info']:
            data_pdf[time_col] = data_pdf[time_col].apply(pd.Timestamp)
    except Exception as e:
        logger.warning(f'{ds_id}-{ds_pid}: {data_file} has malformed datetime columns format. Cannot parse file. Error {e}')
        yield None

    data_pdf = data_pdf[data_pdf[time_primary_var] >= start_date]
    if end_date:
        data_pdf = data_pdf[data_pdf[time_primary_var] <= end_date]

    if data_pdf.shape[0] == 0:
        logger.debug(f'{ds_id}-{ds_pid}: Data file {data_file} does not have timestamps matching the specfied query. Skipping.')
        yield None

    # Next, get variable info
    col_var_info, file_var_list, has_multiple_sensors_cols = _get_queried_variable_info(header_store.get('variable_column_info', {}), observed_property)

    vars_in_file = []
    for header_var_name, var_info in col_var_info.items():
        rf_var_name = var_info.get('variable')
        if rf_var_name not in file_var_list:
            continue
        if header_var_name not in data_pdf.columns:
            logger.info(f'{ds_id}-{ds_pid}: Data file {data_file} header variable {header_var_name} does not have corresponding column name.')
            continue
        vars_in_file.append(rf_var_name)

    # Check if the data file has any of the specified observed properties. If not, skip
    if not vars_in_file:
        logger.debug(f'{ds_id}-{ds_pid}: Data file {data_file} does not have queried observed properties. Skipping.')
        yield None

    # If locations terms are in the column names (e.g., Site_ID, Depth, etc),
    #   then do a little work to separate out the datastreams

    # Get the location terms that are in the columns
    col_loc_terms = list(header_store.get('loc_column_info', []).keys())

    # set up to combo rows for individual locations
    column_term_for_combo = []

    # If Site_ID in column vars, then add, otherwise if lat / lon in col vars, add those
    if id_term and id_in_col and HydroRFTerms.loc_id.value in col_loc_terms:
        column_term_for_combo.append(HydroRFTerms.loc_id.value)
    elif id_term and id_in_col:
        logger.info(f'{ds_id}-{ds_pid}: Data file {data_file} does not match expected format of other data files. Skipping.')
        yield None
    elif id_term is None and id_in_col and HydroRFTerms.lat.value in col_loc_terms and HydroRFTerms.long.value in col_loc_terms:
        # This could be a problem if lat / long info is also in the header or Install Methods file
        column_term_for_combo.extend([HydroRFTerms.lat.value, HydroRFTerms.long.value])
    elif id_term is None and id_in_col:
        logger.info(f'{ds_id}-{ds_pid}: Data file {data_file} does not match expected format of other data files. Skipping.')
        yield None

    # If Sensor_ID is a column var, add
    if has_sensor_id and HydroRFTerms.sensor_id.value in col_loc_terms:
        column_term_for_combo.append(HydroRFTerms.sensor_id.value)
    elif has_sensor_id:
        logger.info(f'{ds_id}-{ds_pid}: Data file {data_file} does not match expected format of other data files. Skipping.')
        yield None

    # check for depth / elevation columns -- add to combo if present
    # ToDo: add some handling for validating the reference datum and units
    if HydroRFTerms.depth.value in col_loc_terms:
        column_term_for_combo.append(HydroRFTerms.depth.value)
    if HydroRFTerms.elevation.value in col_loc_terms:
        column_term_for_combo.append(HydroRFTerms.elevation.value)

    # combo
    # add pdf combo info and the combo terms (or maybe this is in the pandas combo stuff)
    if column_term_for_combo:
        combo_pdf = data_pdf[column_term_for_combo].groupby(column_term_for_combo)
        combo_grp_store = combo_pdf.indices
    else:
        combo_grp_store = {'all': data_pdf.index.array}

    # loop thru the variable columns
    for col_name, col_info in col_var_info.items():
        col_var = col_info.get('variable')
        if col_var not in vars_in_file:
            continue

        col_sensor_id = col_info.get('sensor_id')
        col_unit = col_info.get('unit')

        loc_id: Optional[str] = None
        if id_in_header:
            if id_term == HydroRFTerms.loc_id.value:
                loc_id = col_info.get(HydroRFTerms.loc_id.value)
            else:
                lat = col_info.get(HydroRFTerms.lat.value)
                long = col_info.get(HydroRFTerms.long.value)
                loc_id = _make_lat_long_key(lat, long)

        if id_ref_im:
            im_id = col_info.get(HydroRFTerms.install_method_id.value)
            loc_id = im_ref_store.get(im_id)

        if loc_id:
            if f'{ds_id}-{loc_id}' not in mf_locations or loc_id not in loc_store_index.keys():
                continue
        elif not id_in_col and not loc_id:
            logger.info(f'{ds_id}-{ds_pid}: Data file {data_file} variable {col_name}; Could not find location id.')
            continue

        # determine units
        # Assume the unit for the column matches the RF specification
        col_rf_unit = col_unit

        # Get the RF options for the col variable
        col_rf_var_units_store: Dict = HydroRFTerms.variables.value
        col_rf_units_info = col_rf_var_units_store.get(col_var, {})
        col_rf_units = col_rf_units_info.get('rf_options', [])
        col_convert = col_rf_units_info.get('convert')

        # Check if the column unit is in the RF options for the column variable, if not, see if can fuzzy match it.
        if col_unit not in col_rf_units:
            col_rf_unit = UnitHandler().match_unit(col_rf_units, col_unit)

        # if the unit does not match the specification, skip the column
        if not col_rf_unit:
            logger.info(f'{ds_id}-{ds_pid}: Data file {data_file} variable {col_name} unit {col_unit} not match RF specification. Skipping variable.')
            continue

        conversion_factor = UnitHandler().convert_value(col_rf_unit, col_convert)

        # loop thru the unique loc combo loops, if loc info is not in columns, then only one combo -- the full df
        for col_combo, data_row_indices in combo_grp_store.items():

            data_rows: pd.DataFrame = data_pdf.iloc[data_row_indices, ]

            # find the location_id an/or make it.
            if id_in_col:
                if id_term:
                    loc_id = data_rows[HydroRFTerms.loc_id.value].iloc[0]
                else:
                    lat = data_rows[HydroRFTerms.lat.value].iloc[0]
                    long = data_rows[HydroRFTerms.long.value].iloc[0]
                    loc_id = _make_lat_long_key(lat, long)

            # if loc ID was not found OR the loc_id is not in the query OR the loc_id is not in the locations store, skip
            if not loc_id or (loc_id and f'{ds_id}-{loc_id}' not in mf_locations) or loc_id not in loc_store_index.keys():
                continue

            # otherwise, build the results list
            tvp_results = []
            value_errors = 0
            other_errors = 0

            nrows = data_rows.shape[0]
            for idx in range(0, nrows):
                data_row = data_rows.iloc[idx, ]

                dt_col = data_row[time_primary_var]
                dt_col_iso = dt_col.isoformat()
                val_col = data_row[col_name]
                if val_col == HydroRFTerms.missing_value_numeric.value:
                    continue
                # Checking in case the missing str value was used
                elif val_col == HydroRFTerms.missing_value_str.value:
                    continue

                try:
                    # only try conversion if value is not a numeric type b/c don't want to change an integer if the data is an integer so precision is not inflated
                    if not isinstance(val_col, float) or isinstance(val_col, int):
                        value = float(val_col) * conversion_factor
                    else:
                        value = val_col * conversion_factor
                except ValueError:
                    value_errors += 1
                    continue
                except Exception:
                    other_errors += 1
                    continue

                tvp_results.append(TimeValuePair(timestamp=dt_col_iso, value=value))

            for e_count, e_msg in zip([value_errors, other_errors],
                                      ['data value(s) could not be converted to numeric', 'data value(s) had unexpected errors']):
                if e_count > 0:
                    logger.info(f'{ds_id}-{ds_pid}: Data file {data_file} variable {col_name}: {e_count} {e_msg}')

            # Do we want to return anything if no results?? I think not.
            if not tvp_results:
                logger.info(f'{ds_id}-{ds_pid}: Data file {data_file} had no results that were not missing values')
                continue

            # make the monitoring feature object
            mf_info: Optional[locInfo] = locations_store.get(loc_store_index.get(loc_id))

            if not mf_info:
                logger.info(f'{ds_id}-{ds_pid}: Could not find location information for {loc_id}.')
                continue

            depth, elev, sensor_id = None, None, None
            if HydroRFTerms.depth.value in column_term_for_combo:
                depth_val = data_rows[HydroRFTerms.depth.value].iloc[0]
                try:
                    depth = float(depth_val)
                except Exception:
                    depth = None
            if HydroRFTerms.elevation.value in column_term_for_combo:
                elev_val = data_rows[HydroRFTerms.elevation.value].iloc[0]
                try:
                    elev = float(elev_val)
                except Exception:
                    elev = None

            if col_sensor_id:
                sensor_id = col_sensor_id
            elif HydroRFTerms.sensor_id.value in column_term_for_combo:
                sensor_id = data_rows[HydroRFTerms.sensor_id.value].iloc[0]

            additional_data_info = {
                'data_file': data_file,
                'col_name': col_name,
                'depth': depth,
                'elev': elev,
                'sensor_id': sensor_id,
                'col_unit': col_unit,
                'col_rf_unit': col_rf_unit,
                'conversion_factor': conversion_factor}

            mf_obj = _make_mf_object(plugin_access=plugin_access, mf_info=mf_info, additional_data_info=additional_data_info)

            # make the timeseries object
            mvp = MeasurementTimeseriesTVPObservation(
                plugin_access=plugin_access,
                id=f'MeasTimeseriesTVPObs--{ds_id}-{loc_id}',  # FYI: this field is not unique and thus kinda useless
                feature_of_interest_type=FeatureTypeEnum.POINT,
                feature_of_interest=mf_obj,
                result=ResultListTVP(plugin_access=plugin_access, value=tvp_results),
                observed_property=col_var,
                aggregation_duration='NONE',  # Assuming all data is instantaneous b/c the RF doesn't support it
            )

            # Collect a bit more info:
            utc_offset = None
            if mf_info.utc_offset:
                utc_offset = mf_info.utc_offset
            elif HydroRFTerms.utc_offset.value in col_loc_terms:
                utc_offset = data_rows[HydroRFTerms.utc_offset.value].iloc[0]

            if utc_offset:
                try:
                    mvp.utc_offset = int(utc_offset)
                except Exception:
                    pass

            if time_primary_var == HydroRFTerms.date_time_start.value:
                mvp.time_reference_position = TimeMetadataMixin.TIME_REFERENCE_START

            # yield the object
            yield mvp


class ESSDIVEMonitoringFeatureAccess(DataSourcePluginAccess):
    """
    Access for mapping locations for ESS-DIVE datasets following the Hydrological Monitoring Reporting Format to
    :class:`~basin3d.core.models.MonitoringFeature` objects.

    A location is considered unique on location id
    """
    synthesis_model_class = MonitoringFeature

    def list(self, query: QueryMonitoringFeature) -> Iterator:
        """
        List the monitoring features that match the specified query.

        :param query: a :class:`~basin3d.core.schema.query.QueryMonitoringFeature` object
        :returns: a generator object that yields :class:`~basin3d.core.models.MonitoringFeature` objects
        """

        synthesis_messages: List[str] = []

        datasets = _ess_dive_datasets_handler()

        if query.feature_type and query.feature_type != 'POINT':
            logger.info(f'Dataset {self.datasource.id} does not yet support feature_type {query.feature_type}')
            yield

        if query.parent_feature:
            logger.info(f'Dataset {self.datasource.id} does not support query by parent_feature')
            yield

        mf_ids = []
        if query.monitoring_feature:
            mf_ids = query.monitoring_feature
        elif query.id:
            mf_ids = [query.id]

        mf_store = _build_locations_store(datasets, mf_ids)

        for mf_info in mf_store.values():
            yield _make_mf_object(self, mf_info, is_mf_query=True)

        StopIteration(synthesis_messages)

    def get(self, query):
        """
        Get a single monitoring feature object based on an identifier, i.e. the query must have id specified.

        :param query: a :class:`~basin3d.core.schema.query.QueryMonitoringFeature` object
        :return: a :class:`~basin3d.core.models.MonitoringFeature` object
        """
        if not query.id:
            return None

        mf_iterator = self.list(query)
        mf_list = [mf for mf in mf_iterator]

        if len(mf_list) == 0:
            return None
        elif len(mf_list) > 1:
            logger.warning(f'More than one MonitoringFeature object for id {query.id} only returning first object.')

        return mf_list[0]


class ESSDIVEMeasurementTimeseriesTVPObservationAccess(DataSourcePluginAccess):
    """
    Access for ESS-DIVE data in the Hydrological Monitoring Reporting Format to
    :class:`~basin3d.core.models.MeasurementTimeseriesTVPObservation` objects.
    """
    synthesis_model_class = MeasurementTimeseriesTVPObservation

    def list(self, query: QueryMeasurementTimeseriesTVP):
        """
        Get the data that matches the specified query

        :param query: a :class:`~basin3d.core.schema.query.QueryMeasurementTimeseriesTVP` object
        :return: a generator object that yields :class:`~basin3d.core.models.MeasurementTimeseriesTVPObservation` objects
        """

        synthesis_messages: List[str] = []

        observed_property = query.observed_property
        start_date = pd.Timestamp(query.start_date)
        end_date: Optional[pd.Timestamp] = None
        if query.end_date:
            end_date = pd.Timestamp(query.end_date)
        mf_locations = query.monitoring_feature
        mf_datasets = _extract_ds_id(mf_locations)

        datasets = _ess_dive_datasets_handler()

        location_store = _build_locations_store(datasets, mf_locations=query.monitoring_feature)

        for ds_key, ds_info in datasets.items():
            ds_id, ds_pid = ds_key

            if ds_id not in mf_datasets:
                continue

            data_dir = ds_info.get('data')
            data_files = _list_csv_files(data_dir)

            if not data_files:
                ds_path = ds_info.get('ds_path')
                logger.warning(f'Dataset {ds_id}-{ds_pid} does not have location or data files in {ds_path}. Cannot find locations info.')
                continue

            # Inspect the first data file to determine how to parse the remaining files
            data_file_init = data_files[0]
            data_header_rows, data_pdf = _read_csv_to_pandas_df(data_dir / data_file_init, columns_only=True)

            # parse the header rows
            header_store = _parse_data_file_header_rows(data_header_rows)
            _check_columns_for_loc_info(data_pdf, header_store)

            if not header_store:
                logger.warning(f'{ds_id}-{ds_pid}: Upon initial data file inspection, {data_file_init} has malformed header. Cannot parse dataset.')
                continue

            # 3 cases:
            #   * One file per sensor (could have multiple variables per sensor)
            #   * One file: a column for each sensor (<var_name>_index)
            #   * One file: Sensor_ID column

            # Also, have to find how to get location:
            #   * Site_ID
            #     -- column name
            #     -- header info
            #   * Lat / Long only
            #     -- column names
            #     -- header info
            # 'loc_header_terms': [], 'loc_header_idx': [], 'loc_column_info': {}, 'variable_column_info': {}, 'location_info': {}

            loc_info = ds_info.get('location_info')
            # keys:
            #   all datasets have: 'im_has_lat_long', 'im_has_site_id', 's_has_lat_long', 'id_term',
            #   only T/F if loc info from data files: 'has_id_in_column, 'has_lat_long_in_header', 'has_lat_long_in_column'
            #   id_term = None if using lat / long

            # Is the location identifier in the header, via IM reference, or columns?

            id_term = loc_info.get('id_term')

            col_loc_terms = list(header_store.get('loc_column_info', {}).keys())
            header_loc_terms = header_store.get('loc_header_terms', [])

            id_in_header = False
            id_in_col = False
            id_ref_im = False
            # if location term is Site_ID, where is it in the data file?
            if id_term == HydroRFTerms.loc_id.value:
                if HydroRFTerms.loc_id.value in header_loc_terms:
                    id_in_header = True
                elif HydroRFTerms.loc_id.value in col_loc_terms:
                    id_in_col = True
                elif ds_info.get('location_info').get('im_has_site_id'):
                    id_ref_im = True
            else:
                if _has_lat_long_terms(header_loc_terms):
                    id_in_header = True
                elif _has_lat_long_terms(col_loc_terms):
                    id_in_col = True
                elif ds_info.get('location_info').get('im_has_lat_long'):
                    id_ref_im = True

            if (id_in_header and id_in_col) or (id_in_header and id_ref_im) or (id_in_col and id_ref_im):
                logger.warning(f'{ds_id}-{ds_pid}: {data_file_init} has {HydroRFTerms.loc_id.value} in multiple places in the initial data file inspection. Cannot parse dataset.')
                continue
            elif all([id_in_header, id_in_col, id_ref_im]):
                logger.warning(f'{ds_id}-{ds_pid}: Could not identify location id in initial data file inspection {data_file_init}. Cannot parse dataset.')

            has_sensor_id = False
            if HydroRFTerms.sensor_id.value in col_loc_terms:
                has_sensor_id = True

            # if the location id is referenced via the InstallationMethods file -- parse it and return the dict.
            im_ref_store = {}
            if id_ref_im:
                loc_dir = ds_info.get('locations')
                install_methods_file = ds_info.get('install_methods_file')

                if not install_methods_file:
                    logger.warning(f'{ds_id}-{ds_pid}: Location ID is expected to be in the Install Methods file but cannot find it. Skipping dataset')
                    continue

                im_header_rows, im_pdf = _read_csv_to_pandas_df(loc_dir / install_methods_file)
                im_pdf_columns = list(im_pdf.columns)

                if (HydroRFTerms.install_method_id.value not in im_pdf_columns and
                        (HydroRFTerms.loc_id.value not in im_pdf_columns or
                         HydroRFTerms.lat.value not in im_pdf_columns and HydroRFTerms.long.value not in im_pdf_columns)):
                    logger.info(f'{ds_id}-{ds_pid} Installation Methods file does not have required fields. Cannot parse dataset.')
                    continue

                im_nrows = im_pdf.shape[0]
                for idx in range(0, im_nrows):
                    im_row = im_pdf.iloc[idx, ]
                    im_id = im_row[HydroRFTerms.install_method_id.value]
                    if HydroRFTerms.loc_id.value in im_pdf_columns:
                        im_loc_id = im_row[HydroRFTerms.loc_id.value]
                    else:
                        lat = im_row[HydroRFTerms.lat.value]
                        long = im_row[HydroRFTerms.long.value]
                        im_loc_id = _make_lat_long_key(lat, long)
                    im_ref_store.update({im_id: im_loc_id})

            # loc_id (no ds_prefix): location_store_key
            loc_store_idx = _build_loc_store_index(ds_id, location_store)

            for data_file in data_files:
                meas_tvp_obj_gen = _parse_data_file(
                    plugin_access=self, data_dir=data_dir, data_file=data_file, observed_property=observed_property,
                    mf_locations=mf_locations, start_date=start_date, end_date=end_date, id_term=id_term,
                    id_in_header=id_in_header, id_in_col=id_in_col, id_ref_im=id_ref_im, has_sensor_id=has_sensor_id,
                    ds_id=ds_id, ds_pid=ds_pid, im_ref_store=im_ref_store,
                    loc_store_index=loc_store_idx, locations_store=location_store)
                for mtvpo in meas_tvp_obj_gen:
                    if mtvpo:
                        yield mtvpo

        StopIteration(synthesis_messages)


@basin3d_plugin
class ESSDIVEDataSourcePlugin(DataSourcePluginPoint):
    title = 'ESS-DIVE Hydrological Monitoring Reporting Format Data Source Plugin'
    plugin_access_classes = (ESSDIVEMonitoringFeatureAccess, ESSDIVEMeasurementTimeseriesTVPObservationAccess)

    feature_types = ['POINT']

    class DataSourceMeta:
        """
        This is an internal metadata class for defining additional :class:`~basin3d.models.DataSource`
        attributes.

        **Attributes:**
            - *id* - unique id short name
            - *name* - human friendly name (more descriptive)
            - *location* - resource location
            - *id_prefix* - id prefix to make model object ids unique across plugins
            - *credentials_format* - if the data source requires authentication, this is where the
                format of the stored credentials is defined.

        """
        # Data Source attributes
        id = 'ESSDIVE'  # unique id for the datasource
        location = 'https://ess-dive.data.lbl.gov'
        id_prefix = 'ESSDIVE'
        name = 'ESSDIVE'  # Human Friendly Data Source Name



"""

.. currentmodule:: basin3d.core.synthesis

:platform: Unix, Mac
:synopsis: BASIN-3D ``DataSource`` type classes
:module author: Val Hendrix <vhendrix@lbl.gov>
:module author: Danielle Svehla Christianson <dschristianson@lbl.gov>

.. contents:: Contents
    :local:
    :backlinks: top

"""


class SpatialSamplingShapes(object):
    """
    Spatial sampling shape describing a spatial sampling feature

    Controlled CV list as defined by OGC Observation & Measurement GM_Shape.
    """

    #: The shape of a spatially extensive sampling feature which provides a complete sampling domain.
    SHAPE_SOLID = "SOLID"

    #: The shape of a spatially extensive sampling feature which provides a complete sampling domain.
    SHAPE_SURFACE = "SURFACE"

    #: The shape of a spatially extensive sampling feature which provides a complete sampling domain.
    SHAPE_CURVE = "CURVE"

    #: The shape of a spatially extensive sampling feature which provides a complete sampling domain.
    SHAPE_POINT = "POINT"



"""

.. currentmodule:: basin3d.core.synthesis

:platform: Unix, Mac
:synopsis: BASIN-3D ``DataSource`` catalog classes
:module author: Val Hendrix <vhendrix@lbl.gov>
:module author: Danielle Svehla Christianson <dschristianson@lbl.gov>

.. contents:: Contents
    :local:
    :backlinks: top

"""
import csv
from basin3d.core import monitor
from importlib import resources
from inspect import getmodule
import re
from string import whitespace
from typing import Dict, Iterator, List, Optional, Union

from basin3d.core.schema.enum import MAPPING_DELIMITER, NO_MAPPING_TEXT, MappedAttributeEnum, set_mapped_attribute_enum_type
from basin3d.core.models import DataSource, AttributeMapping, ObservedProperty

logger = monitor.get_logger(__name__)


class CatalogException(Exception):
    pass


class CatalogBase:

    def __init__(self, variable_filename: str = 'basin3d_observed_property_vocabulary.csv'):
        self.variable_dir = 'basin3d.data'
        self.variable_filename = variable_filename

    def initialize(self, plugin_list: list):
        """

        :param plugin_list: list of plugins
        :return: tinyDB object
        """
        if not self.is_initialized():
            logger.debug(f"Initializing {self.__class__.__name__} metadata catalog ")

            # initiate db
            self._init_catalog()

            if not plugin_list:
                plugin_list = []

            # generate variable store
            self._gen_variable_store()

            # Load plugins
            for plugin in plugin_list:

                try:
                    # if the plugin doesn't exist then getting the id should fail
                    plugin_id = plugin.get_meta().id
                    logger.info(f"Loading metadata catalog for Plugin {plugin_id}")
                except Exception:
                    logger.error(
                        'Could not retrieve plugin_id. Check that plugin is configured / registered properly.')
                    raise CatalogException

                mapping_filename = f'{plugin_id.lower()}_mapping.csv'
                plugin_module = getmodule(plugin)
                if not plugin_module:
                    logger.error(f'Plugin {plugin_id} cannot be found')
                    raise CatalogException(f'Plugin {plugin_id} cannot be found')

                plugin_package = ".".join(plugin_module.__name__.split(".")[0:-1])
                # checking for the mapping file might occur elsewhere or look different depending on where plugins live
                has_mapping_file = resources.is_resource(plugin_package, mapping_filename)

                if not has_mapping_file:
                    logger.error(f'Plugin {plugin_id} does not have variable mapping file {mapping_filename}')
                    raise CatalogException(f'Plugin {plugin_id} does not have variable mapping file {mapping_filename}')

                datasource = plugin.get_datasource()
                self._process_plugin_attr_mapping(plugin, mapping_filename, datasource)
                self._insert(datasource)

            logger.info(f"Initialized {self.__class__.__name__} metadata catalog ")

    # ---------------------------------
    # Methods to be over-witten by children classes

    def _init_catalog(self):
        """

        :return:
        """
        raise NotImplementedError

    def _insert(self, record):
        """
        Insert the record
        :param record:
        """
        raise NotImplementedError

    def _get_datasource(self, datasource_id) -> Optional[DataSource]:
        """
        Access a single datasource

        :param datasource_id: the datasource_identifier
        :return:
        """
        raise NotImplementedError

    def _get_observed_property(self, basin3d_vocab) -> Optional[ObservedProperty]:
        """
        Access a single observed property variable

        :param basin3d_vocab: the observed property variable identifier
        :return:
        """
        raise NotImplementedError

    def _get_attribute_mapping(self, datasource_id, attr_type, basin3d_vocab, datasource_vocab, **kwargs) -> Optional[AttributeMapping]:
        """
        Access a single attribute mapping

        :param datasource_id:
        :param attr_type:
        :param basin3d_vocab:
        :param datasource_vocab:
        :return:
        """
        raise NotImplementedError

    def is_initialized(self) -> bool:
        """Has the catalog been initialized?"""
        raise NotImplementedError

    def find_observed_property(self, basin3d_vocab) -> Optional[ObservedProperty]:
        """
        Return the :class:`basin3d.models.ObservedProperty` object for the BASIN-3D vocabulary specified.

        :param basin3d_vocab: BASIN-3D vocabulary
        :return: a :class:`basin3d.models.ObservedProperty` object
        """
        raise NotImplementedError

    def find_observed_properties(self, basin3d_vocab=None) -> Iterator[Optional[ObservedProperty]]:
        """
        Report the observed_properties available based on the BASIN-3D vocabularies specified. If no BASIN-3D vocabularies are specified, then return all observed properties available.

        :param basin3d_vocab: list of the BASIN-3D observed properties
        :return: generator that yields :class:`basin3d.models.ObservedProperty` objects
        """
        raise NotImplementedError

    def find_datasource_attribute_mapping(self, datasource_id, attr_type, attr_vocab) -> Optional[AttributeMapping]:
        """
        Find the datasource attribute vocabulary to BASIN-3D mapping given a specific datasource_id, attr_type, and datasource attr_vocab.

        :param: datasource_id: the datasource identifier
        :param: attr_type: attribute type
        :param: datasource_vocab: the datasource attribute vocabulary
        :return: a :class:`basin3d.models.AttributeMapping` object
        """
        raise NotImplementedError

    def find_attribute_mappings(self, datasource, attr_type, attr_vocab, from_basin3d) -> Iterator[AttributeMapping]:
        """
        Find the list of attribute mappings given the specified fields.
        Exact matches are returned (see attr_vocab formats below for BASIN-3D vocab nuances).
        If no fields are specified, all registered attribute mappings will be returned.

        :param datasource_id: the datasource identifier
        :param attr_type: the attribute type
        :param attr_vocab: the attribute vocabulary, the formats are one of the following:
                           1) datasource vocab that is a complete vocab
                           2) BASIN-3D vocab that is a complete vocab for a given attr_type regardless if it is compound or not.
                           3) BASIN-3D vocab that is compound and fully specified for each attr_type either with the complete vocab or with wildcards.
        :param from_basin3d: boolean that says whether the attr_vocab is a BASIN-3D vocabulary. If not, then this a datasource vocabulary.
        :return: generator that yields :class:`basin3d.models.AttributeMapping` objects
        """
        # class extensions should use _validate_fine_attribute_mappings_arguments
        raise NotImplementedError

    # --------------------------------------

    def _gen_variable_store(self):
        """
        Generate a variable store. Loads this from the provided vocabulary CSV
        :return:
        """
        fields = ['basin3d_vocab', 'description', 'categories', 'units']
        # location of mapping file might change.
        with resources.open_text(self.variable_dir, self.variable_filename) as variable_file:

            reader = csv.DictReader(variable_file)

            # For now, force a specific file format; could change later to just require specific field names
            if reader.fieldnames != fields:
                logger.critical(f'{self.variable_filename} is not in correct format. Cannot create catalog.')
                raise CatalogException

            for row in reader:
                basin3d_vocab = row[fields[0]]

                if self._get_observed_property(basin3d_vocab) is not None:
                    logger.warning(f'Duplicate BASIN-3D variable {basin3d_vocab} found. Skipping duplicate.')
                    continue

                categories_list = []
                if row[fields[2]]:
                    categories_str = row[fields[2]]
                    categories_list = categories_str.split(',')
                    categories_list = [category.strip(whitespace) for category in categories_list]

                observed_property_variable = ObservedProperty(
                    basin3d_vocab=basin3d_vocab,
                    full_name=row[fields[1]],
                    categories=categories_list,
                    units=row[fields[3]])

                self._insert(observed_property_variable)

    def _get_attribute_enum(self, str_value, enum_type):
        """
        Get the enum for a given string value
        :param str_value: string value to convert to enum, e.g., Max
        :param enum_type: the enum type, e.g., StatisticEnum
        :return:
        """
        enum_value = None
        try:
            enum_value = getattr(enum_type, str_value)
        except AttributeError:
            pass
        return enum_value

    def _process_plugin_attr_mapping(self, plugin, filename: str, datasource: DataSource):
        """

        :param plugin:
        :param filename:
        :param datasource:
        :return:
        """
        fields = ['attr_type', 'basin3d_vocab', 'datasource_vocab', 'datasource_desc']

        plugin_id = plugin.get_meta().id
        plugin_module = getmodule(plugin)
        if not plugin_module:
            logger.error(f'Plugin {plugin_id} cannot be found')
            raise CatalogException(f'Plugin {plugin_id} cannot be found')

        plugin_package = ".".join(plugin_module.__name__.split(".")[0:-1])

        with resources.open_text(plugin_package, filename) as attr_file:
            logger.debug(f"Mapping file {filename} for plugin package {plugin_package}")
            reader = csv.DictReader(attr_file)

            # For now, force a specific file format; could change later to just require specific field names
            if reader.fieldnames != fields:
                logger.critical(
                    f'Plugin {plugin_id}: {filename} is not in correct format. Cannot create catalog.')
                raise CatalogException(
                    f'Plugin {plugin_id}: {filename} is not in correct format. Cannot create catalog.')

            # loop thru the file
            for row in reader:
                attr_type = row[fields[0]]
                basin3d_vocab = row[fields[1]]
                datasource_vocab = row[fields[2]]
                datasource_desc = row[fields[3]]

                # ToDo: add some validation if Cat does not have it in her reader
                if self._get_attribute_mapping(datasource.id, attr_type, basin3d_vocab, datasource_vocab) is not None:
                    logger.warning(
                        f'Plugin {plugin_id}: Duplicate BASIN-3D attr detected. Cannot handle duplicate mappings yet. '
                        f'Skipping datasource attribute {datasource_vocab}.')

                # build the basin3d_description from objects (e.g., for OBSERVED_PROPERTY) or enums
                basin3d_desc = []
                for a_type, b3d_vocab in zip(attr_type.split(MAPPING_DELIMITER), basin3d_vocab.split(MAPPING_DELIMITER)):
                    valid_type_vocab = True

                    b3d_attr_type = None
                    try:
                        b3d_attr_type = getattr(MappedAttributeEnum, a_type)
                        if b3d_attr_type == MappedAttributeEnum.OBSERVED_PROPERTY:
                            b3d_desc = self._get_observed_property(b3d_vocab)
                        else:
                            b3d_enum_type = set_mapped_attribute_enum_type(b3d_attr_type)
                            b3d_desc = self._get_attribute_enum(b3d_vocab, b3d_enum_type)
                    except AttributeError:
                        logger.warning(f'Datasource {datasource.id}: Attribute type {a_type} is not supported. Skipping mapping.')
                        valid_type_vocab = False
                        break

                    if not b3d_desc:
                        logger.warning(f'Datasource {datasource.id}: basin3d_vocab {b3d_vocab} for attr_type {a_type} is not a valid BASIN-3D vocabulary. Skipping attribute mapping.')
                        valid_type_vocab = False
                        break

                    basin3d_desc.append(b3d_desc)

                if not valid_type_vocab:
                    continue

                attr_mapping = AttributeMapping(
                    attr_type=attr_type,
                    basin3d_vocab=basin3d_vocab,
                    basin3d_desc=basin3d_desc,
                    datasource_vocab=datasource_vocab,
                    datasource_desc=datasource_desc,
                    datasource=datasource)

                self._insert(attr_mapping)
                logger.debug(f"{datasource.id}: Mapped {attr_type} {datasource_vocab} to {basin3d_vocab}")

    def _validate_find_attribute_mappings_arguments(self, datasource_id: str = None, attr_type: str = None):
        """
        Validate arguments for method find_attribute_mappings

        :param datasource_id: datasource ID
        :param attr_type: Attribute Type
        """
        if datasource_id and self._get_datasource(datasource_id) is None:
            msg = f'Specified datasource_id: "{datasource_id}" has not been registered.'
            logger.critical(msg)
            raise CatalogException(msg)

        if attr_type and attr_type not in MappedAttributeEnum.values():
            msg = f'"{attr_type}" is not an attribute type supported by BASIN-3D.'
            logger.critical(msg)
            raise CatalogException(msg)


class CatalogTinyDb(CatalogBase):

    def __init__(self, variable_filename: str = 'basin3d_observed_property_vocabulary.csv'):
        super().__init__(variable_filename)

        self.in_memory_db = None
        self.in_memory_db_attr = None
        self._observed_properties: Dict[str, ObservedProperty] = {}
        self._attribute_mappings: Dict[str, AttributeMapping] = {}
        self._datasources: Dict[str, DataSource] = {}

    def is_initialized(self) -> bool:
        """Has the catalog been initialized?"""
        return self.in_memory_db is not None

    def _get_datasource(self, datasource_id) -> Optional[DataSource]:
        """
        Access a datasource via the datasource_id
        :param datasource_id: str, the datasource id
        :return: a :class:`basin3d.models.DataSource` object
        """
        return self._datasources.get(datasource_id, None)

    def _get_observed_property(self, basin3d_vocab) -> Optional[ObservedProperty]:
        """
        Access a single observed property

        :param basin3d_vocab: str, the observed property identifier
        :return: an :class:`basin3d.models.ObservedProperty` object
        """
        return self._observed_properties.get(basin3d_vocab, None)

    def _get_attribute_mapping(self, datasource_id, attr_type, basin3d_vocab, datasource_vocab, **kwargs) -> Optional[AttributeMapping]:
        """
        Access a single attribute mapping

        :param datasource_id: str, datasource identifier
        :param attr_type: str, attribute type
        :param basin3d_vocab: str, BASIN-3D vocabulary
        :param datasource_vocab: str, datasource vocabulary
        :return: a :class:`basin3d.models.AttributeMapping` object
        """
        return self._attribute_mappings.get(f'{datasource_id}-{attr_type}-{basin3d_vocab}-{datasource_vocab}', None)

    def find_observed_property(self, basin3d_vocab: str) -> Optional[ObservedProperty]:
        """
        Return the :class:`basin3d.models.ObservedProperty` object for the BASIN-3D vocabulary specified.

        :param basin3d_vocab: BASIN-3D vocabulary
        :return: a :class:`basin3d.models.ObservedProperty` object
        """

        if not self._observed_properties:
            msg = "Variable Store has not been initialized."
            logger.critical(msg)
            raise CatalogException(msg)

        return self._get_observed_property(basin3d_vocab)

    def find_observed_properties(self, basin3d_vocab: Optional[List[str]] = None) -> Iterator[Optional[ObservedProperty]]:
        """
        Report the observed_properties available based on the BASIN-3D vocabularies specified. If no BASIN-3D vocabularies are specified, then return all observed properties available.

        :param basin3d_vocab: list of the BASIN-3D observed properties
        :return: generator that yields :class:`basin3d.models.ObservedProperty` objects
        """

        if not self._observed_properties:
            msg = "Variable Store has not been initialized."
            logger.critical(msg)
            raise CatalogException(msg)

        opv: Optional[ObservedProperty]
        if basin3d_vocab is None:
            for opv in self._observed_properties.values():
                yield opv
        else:
            for b3d_vocab in basin3d_vocab:
                opv = self._get_observed_property(b3d_vocab)
                if opv is not None:
                    yield opv
                else:
                    logger.warning(f'BASIN-3D does not support variable {b3d_vocab}')

    def find_datasource_attribute_mapping(self, datasource_id: str, attr_type: str, datasource_vocab: str) -> Optional[AttributeMapping]:
        """
        Find the datasource attribute vocabulary to BASIN-3D mapping given a specific datasource_id, attr_type, and datasource attr_vocab.

        :param: datasource_id: the datasource identifier
        :param: attr_type: attribute type
        :param: datasource_vocab: the datasource attribute vocabulary
        :return: a :class:`basin3d.models.AttributeMapping` object
        """
        if self.in_memory_db_attr is None:
            msg = 'Attribute Store has not been initialized.'
            logger.critical(msg)
            raise CatalogException(msg)

        from tinydb import Query
        query = Query()

        # attr_type is search to accommodate compound mappings
        results = self.in_memory_db_attr.search(
            (query.datasource_vocab == datasource_vocab) & (query.datasource_id == datasource_id) & (query.attr_type.search(attr_type)))

        # Only one result should be returned. If so return it and all is good.
        if len(results) == 1:
            return self._get_attribute_mapping(**results[0])
        # If not, deal with results:
        # Case 1: more than one mapping for the datasource vocabulary. This should not happen. Raise an exception.
        elif len(results) > 1:
            error_msg = (f'More than one attribute mapping found for datasource vocab: "{datasource_vocab}" '
                         f'in datasource: "{datasource_id}". This should never happen.')
            logger.critical(error_msg)
            raise CatalogException(error_msg)

        # Case 2: no mapping was found. Return an empty AttributeMapping object
        msg = f'No mapping was found for datasource vocab: "{datasource_vocab}" in datasource: "{datasource_id}".'
        datasource = self._get_datasource(datasource_id)

        if datasource is None:
            datasource = DataSource()
            msg = f'No datasource was found for id "{datasource_id}".'
            logger.warning(msg)

        return AttributeMapping(attr_type=attr_type, basin3d_vocab=NO_MAPPING_TEXT, basin3d_desc=[],
                                datasource_vocab=datasource_vocab, datasource_desc=msg, datasource=datasource)

    def find_attribute_mappings(self, datasource_id: str = None, attr_type: str = None, attr_vocab: Union[str, List] = None,
                                from_basin3d: bool = False) -> Iterator[AttributeMapping]:
        """
        Find the list of attribute mappings given the specified fields.
        Exact matches are returned (see attr_vocab formats below for BASIN-3D vocab nuances).
        If no fields are specified, all registered attribute mappings will be returned.

        :param datasource_id: the datasource identifier
        :param attr_type: the attribute type
        :param attr_vocab: the attribute vocabulary, the formats are one of the following:
                           1) datasource vocab that is a complete vocab
                           2) BASIN-3D vocab that is a complete vocab for a given attr_type regardless if it is compound or not.
                           3) BASIN-3D vocab that is compound and fully specified for each attr_type either with the complete vocab or with wildcards.
        :param from_basin3d: boolean that says whether the attr_vocab is a BASIN-3D vocabulary. If not, then this a datasource vocabulary.
        :return: generator that yields :class:`basin3d.models.AttributeMapping` objects
        """
        catalog_messages: List[str] = []

        self._validate_find_attribute_mappings_arguments(datasource_id, attr_type)

        if self.in_memory_db_attr is None:
            msg = 'Attribute Store has not been initialized.'
            logger.critical(msg)
            raise CatalogException(msg)

        from tinydb import Query
        query = Query()

        # Set the attribute vocab list to track mappings not found
        specified_attr_vocab_not_found = []
        if attr_vocab:
            if isinstance(attr_vocab, str):
                attr_vocab = [attr_vocab]
            elif not isinstance(attr_vocab, List):
                raise CatalogException("attr_vocab must be a str or list")
            specified_attr_vocab_not_found = attr_vocab.copy()

        # Function for TinyDB search
        def is_in(x, attr_vocabs=attr_vocab, is_from_basin3d=from_basin3d):
            for a_vocab in attr_vocabs:
                # if attr_vocabs is datasource vocab(s)
                # OR attr_vocabs is BASIN-3D vocab(s) AND they have the compound mapping delimiter (i.e., it has : and possibly wildcards)
                # AND it is a fullmatch
                if (not is_from_basin3d or (is_from_basin3d and MAPPING_DELIMITER in a_vocab)) and re.fullmatch(a_vocab, x):
                    return True
                # if attr_vocabs is BASIN-3D vocab(s) and they do not have the compound mapping delimiter,
                # then match exactly for the individual attr_type components.
                # e.g. attr_vocabs = ['MAX', 'MIN'] and the mapping is compound e.g. x = 'Al:WATER:MAX'
                # re.search does not work here b/c of a case like OBSERVED_PROPERPTY variables TEMP:AIR and SONIC_TEMP:AIR,
                # in which a attr_vocab of ['TEMP'] would return both instead on just TEMP:AIR.
                # A general search would need to be handled differently.
                elif is_from_basin3d and MAPPING_DELIMITER not in a_vocab:
                    for x_vocab in x.split(MAPPING_DELIMITER):
                        if x_vocab == a_vocab:
                            return True
            return False

        # If no query parameters --> get all mapped attributes back for all registered plugins
        if not datasource_id and not attr_vocab and not attr_type:
            # return all mapped attributes possible possible
            for attr_mapping in self._attribute_mappings.values():
                yield attr_mapping
        # Otherwise search depends on the set of parameters provided
        else:
            if not datasource_id:
                if attr_type and attr_vocab and from_basin3d:
                    results = self.in_memory_db_attr.search((query.basin3d_vocab.test(is_in)) & (query.attr_type.search(attr_type)))
                elif attr_type and attr_vocab:
                    results = self.in_memory_db_attr.search((query.datasource_vocab.test(is_in)) & (query.attr_type.search(attr_type)))
                elif not attr_type and from_basin3d:
                    results = self.in_memory_db_attr.search(query.basin3d_vocab.test(is_in))
                elif not attr_type:
                    results = self.in_memory_db_attr.search(query.datasource_vocab.test(is_in))
                else:
                    results = self.in_memory_db_attr.search(query.attr_type.search(attr_type))

            # Returns all possible attributes for a data source
            elif not attr_type and not attr_vocab:
                results = self.in_memory_db_attr.search(query.datasource_id == datasource_id)

            # Returns all attributes for a data source and attribute type
            elif not attr_vocab:
                results = self.in_memory_db_attr.search((query.datasource_id == datasource_id) & (query.attr_type.search(attr_type)))

            # Return mappings if only the datasource and attr_vocab are specified.
            elif not attr_type:
                if from_basin3d:
                    results = self.in_memory_db_attr.search((query.basin3d_vocab.test(is_in)) & (query.datasource_id == datasource_id))
                else:
                    results = self.in_memory_db_attr.search((query.datasource_vocab.test(is_in)) & (query.datasource_id == datasource_id))

            # Finally, all parameters specified:
            # Convert from BASIN-3D to DataSource variable name
            elif from_basin3d:
                results = self.in_memory_db_attr.search(
                    (query.basin3d_vocab.test(is_in)) & (query.datasource_id == datasource_id) & (query.attr_type.search(attr_type)))

            # Convert from DataSource variable name to BASIN-3D
            else:
                results = self.in_memory_db_attr.search(
                    (query.datasource_vocab.test(is_in)) & (query.datasource_id == datasource_id) & (query.attr_type.search(attr_type)))

            # Yield the results
            attr_map: Optional[AttributeMapping]
            for r in results:
                attr_map = self._get_attribute_mapping(**r)

                # if AttributeMapping is not found: THIS SHOULD NEVER HAPPEN.
                if attr_map is None:
                    logger.error('AttributeMapping not found given database search results. THIS SHOULD NEVER HAPPPEN.')
                    continue

                # If attr_vocab was specified, track what was found for the datasource
                if attr_vocab:
                    vocabs = [attr_map.datasource_vocab]
                    if from_basin3d:
                        vocabs = attr_map.basin3d_vocab.split(MAPPING_DELIMITER)

                    # look for it in the specified vocabs, remove it from the not found list
                    for vocab in vocabs:
                        try:
                            idx = specified_attr_vocab_not_found.index(vocab)
                            specified_attr_vocab_not_found.pop(idx)
                        except ValueError:
                            pass

                # yield the AttributeMapping
                yield attr_map

            # If any specified attr_vocabs were not found, add info messages
            if specified_attr_vocab_not_found:
                attr_vocab_text = ', '.join(specified_attr_vocab_not_found)
                attr_vocab_source = 'datasource'
                if from_basin3d:
                    attr_vocab_source = 'BASIN-3D'
                if not results:
                    # if no results,
                    msg = (f'No attribute mappings found for specified parameters: datasource id = "{datasource_id}", '
                           f'attribute type = "{attr_type}", {attr_vocab_source} vocabularies: {attr_vocab_text}.')
                else:
                    msg = (f'No attribute mappings found for the following {attr_vocab_source} vocabularies: {attr_vocab_text}. '
                           f'Note: specified datasource id = {datasource_id} and attribute type = {attr_type}')
                logger.info(msg)
                catalog_messages.append(msg)

            return StopIteration(catalog_messages)

    def _init_catalog(self):
        """
        Initialize the catalog database

        :return:
        """
        # Note: op are held in dictionary in this TinyDB
        from tinydb import TinyDB
        from tinydb.storages import MemoryStorage
        self.in_memory_db = TinyDB(storage=MemoryStorage)
        self.in_memory_db_attr = self.in_memory_db.table('attr')
        self.in_memory_db_attr.truncate()

    def _insert(self, record):
        """
        :param record:
        """
        if self.in_memory_db is not None:
            if isinstance(record, ObservedProperty):
                self._observed_properties[record.basin3d_vocab] = record
            elif isinstance(record, AttributeMapping):
                key = f'{record.datasource.id}-{record.attr_type}-{record.basin3d_vocab}-{record.datasource_vocab}'
                logger.debug(key)
                self._attribute_mappings[key] = record
                self.in_memory_db_attr.insert({'datasource_id': record.datasource.id,
                                               'attr_type': record.attr_type,
                                               'basin3d_vocab': record.basin3d_vocab,
                                               'basin3d_desc': record.basin3d_desc,
                                               'datasource_vocab': record.datasource_vocab,
                                               'datasource_desc': record.datasource_desc, })
            elif isinstance(record, DataSource):
                self._datasources[record.id] = record
        else:
            msg = 'Could not insert record. Catalog not initialized.'
            logger.critical(msg)
            raise CatalogException(msg)



"""

.. currentmodule:: basin3d.core.schema.enum

:platform: Unix, Mac
:synopsis: BASIN-3D Enumeration Schema
:module author: Val Hendrix <vhendrix@lbl.gov>
:module author: Danielle Svehla Christianson <dschristianson@lbl.gov>

.. contents:: Contents
    :local:
    :backlinks: top


"""
from enum import Enum

from basin3d.core.types import SpatialSamplingShapes

MAPPING_DELIMITER = ':'
NO_MAPPING_TEXT = 'NOT_SUPPORTED'


class BaseEnum(Enum):
    """Base Enumeration Class that adds some helper methods"""

    @classmethod
    def values(cls):
        role_names = [member.value for role, member in cls.__members__.items()]
        return role_names

    @classmethod
    def names(cls):
        return cls._member_names_


class MappedAttributeEnum(str, BaseEnum):
    """
    Enumeration for mapped attributes
    """
    OBSERVED_PROPERTY = 'OBSERVED_PROPERTY'
    AGGREGATION_DURATION = 'AGGREGATION_DURATION'
    RESULT_QUALITY = 'RESULT_QUALITY'
    SAMPLING_MEDIUM = 'SAMPLING_MEDIUM'
    STATISTIC = 'STATISTIC'


class TimeFrequencyEnum(str, BaseEnum):
    """
    Enumeration for time frequencies
    """
    # ToDo: Check OGC for correct term (i.e., replace TimeFrequency)

    YEAR = "YEAR"
    MONTH = "MONTH"
    DAY = "DAY"
    HOUR = "HOUR"
    MINUTE = "MINUTE"
    SECOND = "SECOND"
    NONE = "NONE"
    NOT_SUPPORTED = NO_MAPPING_TEXT


class AggregationDurationEnum(str, BaseEnum):
    """
    Aggregation Duration enums
    """
    #: Observations aggregated by year
    YEAR = TimeFrequencyEnum.YEAR.value

    #: Observations aggregated by month
    MONTH = TimeFrequencyEnum.MONTH.value

    #: Observations aggregated by day
    DAY = TimeFrequencyEnum.DAY.value

    #: Observations aggregated by hour
    HOUR = TimeFrequencyEnum.HOUR.value

    #: Observations aggregated by minute
    MINUTE = TimeFrequencyEnum.MINUTE.value

    #: Observations aggregated by second
    SECOND = TimeFrequencyEnum.SECOND.value

    #: Observations aggregated by no standard frequency, used for instantaneous values
    NONE = TimeFrequencyEnum.NONE.value


class FeatureTypeEnum(str, BaseEnum):
    """Enumeration for Feature Types"""
    REGION = "REGION"
    SUBREGION = "SUBREGION"
    BASIN = "BASIN"
    SUBBASIN = "SUBBASIN"
    WATERSHED = "WATERSHED"
    SUBWATERSHED = "SUBWATERSHED"
    SITE = "SITE"
    PLOT = "PLOT"
    HORIZONTAL_PATH = "HORIZONTAL_PATH"
    VERTICAL_PATH = "VERTICAL_PATH"
    POINT = "POINT"


FEATURE_SHAPE_TYPES = {
    SpatialSamplingShapes.SHAPE_POINT: [FeatureTypeEnum.POINT],
    SpatialSamplingShapes.SHAPE_CURVE: [FeatureTypeEnum.HORIZONTAL_PATH, FeatureTypeEnum.VERTICAL_PATH],
    SpatialSamplingShapes.SHAPE_SURFACE: [FeatureTypeEnum.REGION, FeatureTypeEnum.SUBREGION, FeatureTypeEnum.BASIN,
                                          FeatureTypeEnum.SUBBASIN,
                                          FeatureTypeEnum.WATERSHED,
                                          FeatureTypeEnum.SUBWATERSHED, FeatureTypeEnum.SITE, FeatureTypeEnum.PLOT],
    SpatialSamplingShapes.SHAPE_SOLID: []
}


class ResultQualityEnum(str, BaseEnum):
    """Enumeration for Result Quality"""

    #: The result is raw or unchecked for quality. Synonyms: Unchecked, Preliminary, No QC
    UNVALIDATED = "UNVALIDATED"

    #: The result has been checked for quality and no issues identified. Synonyms: Checked, Accepted, Pass, OK, Good
    VALIDATED = "VALIDATED"

    #: The result is identified as poor quality. Synonyms: Poor, Bad, Unaccepted
    REJECTED = "REJECTED"

    #: The result's quality is suspect. Synonyms: Questionable, Doubtful, Spike/Noise, Flagged
    SUSPECTED = "SUSPECTED"

    #: The result is estimated. Synonyms: Interpolated, Modeled.
    ESTIMATED = "ESTIMATED"

    #: The quality type is not supported
    NOT_SUPPORTED = NO_MAPPING_TEXT


class StatisticEnum(str, BaseEnum):
    """Enumeration for Statistics"""
    INSTANT = "INSTANT"
    MEAN = "MEAN"
    MIN = "MIN"
    MAX = "MAX"
    TOTAL = "TOTAL"
    NOT_SUPPORTED = NO_MAPPING_TEXT


class SamplingMediumEnum(str, BaseEnum):
    """
    Types of sampling mediums for Observed Properties
    """
    SOLID_PHASE = "SOLID_PHASE"
    WATER = "WATER"
    GAS = "GAS"
    OTHER = "OTHER"
    NOT_APPLICABLE = "NOT_APPLICABLE"
    NOT_SUPPORTED = NO_MAPPING_TEXT


class MessageLevelEnum(str, BaseEnum):
    """Enumeration of Message Levels"""

    WARN = "WARN"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


def set_mapped_attribute_enum_type(attr_type: str):
    """
    Return the enum type for the specified MappedAttributeEnum.
    :param attr_type: MappedAttributeEnum
    :return: the type's Enum
    """
    if attr_type == MappedAttributeEnum.AGGREGATION_DURATION:
        return AggregationDurationEnum
    elif attr_type == MappedAttributeEnum.RESULT_QUALITY:
        return ResultQualityEnum
    elif attr_type == MappedAttributeEnum.SAMPLING_MEDIUM:
        return SamplingMediumEnum
    elif attr_type == MappedAttributeEnum.STATISTIC:
        return StatisticEnum
    else:
        return None
